import json
import logging
import os
import subprocess
import time

import psycopg

logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")
from typing import Any, Dict, Optional, Tuple


def start_docker_pg() -> None:
    """
    Starts a PostgreSQL database in a Docker container.

    This function checks if Docker is installed, and if so, starts a local PostgreSQL
    database in a Docker container. It configures the database with default settings
    and provides connection information upon successful startup.

    The function uses environment variable PGPASSWORD if available, otherwise
    defaults to 'dbos' as the database password.

    Returns:
        None

    Raises:
        Exception: If there is an error starting the Docker container or if the
                   PostgreSQL service does not become available within the timeout period.
    """

    logging.info("Attempting to create a Docker Postgres container...")
    has_docker = check_docker_installed()

    pool_config = {
        "host": "localhost",
        "port": 5432,
        "password": os.environ.get("PGPASSWORD", "dbos"),
        "user": "postgres",
        "database": "postgres",
        "connect_timeout": 2,
    }

    # If Docker is installed, start a local Docker based Postgres
    if has_docker:
        start_docker_postgres(pool_config)
        logging.info(
            f"Postgres available at postgresql://postgres:{pool_config['password']}@{pool_config['host']}:{pool_config['port']}"
        )
    else:
        logging.warning("Docker not detected locally")


def check_db_connectivity(config: Dict[str, Any]) -> Optional[Exception]:
    conn = None
    try:
        conn = psycopg.connect(
            host=config["host"],
            port=config["port"],
            user=config["user"],
            password=config["password"],
            dbname=config["database"],
            connect_timeout=config.get("connect_timeout", 30),
        )
        cursor = conn.cursor()
        cursor.execute("SELECT 1;")
        cursor.close()
        return None
    except Exception as error:
        return error
    finally:
        if conn is not None:
            conn.close()


def exec_sync(cmd: str) -> Tuple[str, str]:
    result = subprocess.run(cmd, shell=True, text=True, capture_output=True, check=True)
    return result.stdout, result.stderr


def start_docker_postgres(pool_config: Dict[str, Any]) -> bool:
    logging.info("Starting a Postgres Docker container...")
    container_name = "dbos-db"
    pg_data = "/var/lib/postgresql/data"
    image_name = "pgvector/pgvector:pg16"

    try:
        # Check if the container already exists
        try:
            result = subprocess.run(
                f"docker inspect {container_name}",
                shell=True,
                text=True,
                capture_output=True,
            )

            if result.returncode == 0:
                # Container exists, check its status
                container_info = json.loads(result.stdout)
                status = container_info[0]["State"]["Status"]

                if status == "running":
                    logging.info(f"Container '{container_name}' is already running.")
                    return True
                elif status == "exited":
                    subprocess.run(
                        f"docker start {container_name}", shell=True, check=True
                    )
                    logging.info(
                        f"Container '{container_name}' was stopped and has been restarted."
                    )
                    return True
        except (
            subprocess.CalledProcessError,
            json.JSONDecodeError,
            KeyError,
            IndexError,
        ):
            # Container doesn't exist or error parsing, proceed with creation
            pass

        # Check if the image exists locally
        result = subprocess.run(
            f"docker images -q {image_name}", shell=True, text=True, capture_output=True
        )

        if not result.stdout.strip():
            logging.info(f"Pulling Docker image {image_name}...")
            subprocess.run(f"docker pull {image_name}", shell=True, check=True)

        # Create and start the container
        cmd = [
            "docker run",
            "-d",
            f"--name {container_name}",
            f"-e POSTGRES_PASSWORD={pool_config['password']}",
            f"-e PGDATA={pg_data}",
            f"-p {pool_config['port']}:5432",
            f"-v {pg_data}:{pg_data}",
            "--rm",
            image_name,
        ]

        result = subprocess.run(
            " ".join(cmd), shell=True, text=True, capture_output=True, check=True
        )

        container_id = result.stdout.strip()
        logging.info(f"Created container: {container_id}")

    except subprocess.CalledProcessError as e:
        raise Exception(f"Docker command error: {e.stderr if e.stderr else str(e)}")

    # Wait for PostgreSQL to be ready
    attempts = 30
    while attempts > 0:
        if attempts % 5 == 0:
            logging.info("Waiting for Postgres Docker container to start...")

        if check_db_connectivity(pool_config) is None:
            return True

        attempts -= 1
        time.sleep(1)

    raise Exception(
        f"Failed to start Docker container: Container {container_name} did not start in time."
    )


def check_docker_installed() -> bool:
    """
    Check if Docker is installed and running using the Docker CLI.

    Returns:
        bool: True if Docker is installed and running, False otherwise.
    """
    try:
        result = subprocess.run(
            "docker version --format json", shell=True, capture_output=True, text=True
        )
        return result.returncode == 0
    except Exception:
        return False


def stop_docker_pg() -> None:
    """
    Stops the Docker Postgres container.

    Returns:
        bool: True if the container was successfully stopped, False if it wasn't running

    Raises:
        Exception: If there was an error stopping the container
    """
    logger = logging.getLogger()
    container_name = "dbos-db"
    try:
        logger.info(f"Stopping Docker Postgres container {container_name}...")

        # Check if container exists
        result = subprocess.run(
            f"docker inspect {container_name}",
            shell=True,
            text=True,
            capture_output=True,
        )

        if result.returncode == 0:
            # Container exists, check its status
            try:
                container_info = json.loads(result.stdout)
                status = container_info[0]["State"]["Status"]

                if status == "running":
                    subprocess.run(
                        f"docker stop {container_name}", shell=True, check=True
                    )
                    logger.info(
                        f"Successfully stopped Docker Postgres container {container_name}."
                    )
                else:
                    logger.info(
                        f"Container {container_name} exists but is not running."
                    )
            except (json.JSONDecodeError, KeyError, IndexError) as e:
                logger.error(f"Error parsing container info: {e}")
                raise
        else:
            logger.info(f"Container {container_name} does not exist.")

    except subprocess.CalledProcessError as error:
        error_message = error.stderr if error.stderr else str(error)
        logger.error(f"Failed to stop Docker Postgres container: {error_message}")
        raise
    except Exception as error:
        error_message = str(error)
        logger.error(f"Failed to stop Docker Postgres container: {error_message}")
        raise



import glob
import os
import subprocess
import sys
import time
from typing import Any, Generator, Tuple
from urllib.parse import quote

import pytest
import sqlalchemy as sa
from fastapi import FastAPI
from flask import Flask

from dbos import DBOS, DBOSClient, DBOSConfig
from dbos._schemas.system_database import SystemSchema


@pytest.fixture(scope="session")
def build_wheel() -> str:
    subprocess.check_call(["pdm", "build"])
    wheel_files = glob.glob(os.path.join("dist", "*.whl"))
    assert len(wheel_files) == 1
    return wheel_files[0]


def using_sqlite() -> bool:
    return os.environ.get("DBOS_DATABASE", None) == "SQLITE"


@pytest.fixture()
def skip_with_sqlite() -> None:
    if using_sqlite():
        pytest.skip("Skipping test when testing SQLite")


@pytest.fixture()
def skip_with_sqlite_imprecise_time() -> None:
    if using_sqlite() and sys.version_info < (3, 12):
        pytest.skip(
            "Skipping test when testing SQLite on Python version <3.12 as SQLite lacks ms-precision timestamps"
        )


def default_config() -> DBOSConfig:
    return {
        "name": "test-app",
        "application_database_url": (
            "sqlite:///test.sqlite"
            if using_sqlite()
            else f"postgresql://postgres:{quote(os.environ.get('PGPASSWORD', 'dbos'), safe='')}@localhost:5432/dbostestpy"
        ),
        "system_database_url": (
            "sqlite:///test.sqlite"
            if using_sqlite()
            else f"postgresql+psycopg://postgres:{quote(os.environ.get('PGPASSWORD', 'dbos'), safe='')}@localhost:5432/dbostestpy_dbos_sys"
        ),
        "enable_otlp": True,
    }


@pytest.fixture()
def config() -> DBOSConfig:
    return default_config()


@pytest.fixture(scope="session")
def db_engine() -> Generator[sa.Engine, Any, None]:
    cfg = default_config()
    assert cfg["system_database_url"] is not None
    if using_sqlite():
        engine = sa.create_engine(cfg["system_database_url"])
        yield engine
    else:
        engine = sa.create_engine(
            sa.make_url(cfg["system_database_url"]).set(
                drivername="postgresql+psycopg",
                database="postgres",
            ),
            connect_args={
                "connect_timeout": 30,
            },
        )
        yield engine
    engine.dispose()


@pytest.fixture()
def cleanup_test_databases(config: DBOSConfig, db_engine: sa.Engine) -> None:
    assert config["application_database_url"] is not None
    assert config["system_database_url"] is not None

    if using_sqlite():
        # For SQLite, delete the database files
        # Extract file path from SQLite URL
        parsed_url = sa.make_url(config["system_database_url"])
        db_path = parsed_url.database
        assert db_path is not None

        # Remove the database files if they exist
        if os.path.exists(db_path):
            os.remove(db_path)
    else:
        # For PostgreSQL, drop the databases
        app_db_name = sa.make_url(config["application_database_url"]).database
        sys_db_name = sa.make_url(config["system_database_url"]).database

        with db_engine.connect() as connection:
            connection.execution_options(isolation_level="AUTOCOMMIT")
            connection.execute(
                sa.text(f"DROP DATABASE IF EXISTS {app_db_name} WITH (FORCE)")
            )
            connection.execute(
                sa.text(f"DROP DATABASE IF EXISTS {sys_db_name} WITH (FORCE)")
            )

    # Clean up environment variables
    os.environ.pop("DBOS__VMID") if "DBOS__VMID" in os.environ else None
    os.environ.pop("DBOS__APPVERSION") if "DBOS__APPVERSION" in os.environ else None
    os.environ.pop("DBOS__APPID") if "DBOS__APPID" in os.environ else None


@pytest.fixture()
def dbos(
    config: DBOSConfig, cleanup_test_databases: None
) -> Generator[DBOS, Any, None]:
    DBOS.destroy(destroy_registry=True)

    # This launches for test convenience.
    #    Tests add to running DBOS and then call stuff without adding
    #     launch themselves.
    # If your test is tricky and has a problem with this, use a different
    #   fixture that does not launch.
    dbos = DBOS(config=config)
    DBOS.launch()

    yield dbos
    DBOS.destroy(destroy_registry=True)


@pytest.fixture()
def client(config: DBOSConfig, dbos: DBOS) -> Generator[DBOSClient, Any, None]:
    assert config["application_database_url"] is not None
    assert config["system_database_url"] is not None
    client = DBOSClient(
        application_database_url=config["application_database_url"],
        system_database_url=config["system_database_url"],
    )
    yield client
    client.destroy()


@pytest.fixture()
def dbos_fastapi(
    config: DBOSConfig, cleanup_test_databases: None
) -> Generator[Tuple[DBOS, FastAPI], Any, None]:
    DBOS.destroy(destroy_registry=True)
    app = FastAPI()
    dbos = DBOS(fastapi=app, config=config)

    # This is for test convenience.
    #    Usually fastapi itself does launch, but we are not completing the fastapi lifecycle
    DBOS.launch()

    yield dbos, app
    DBOS.destroy(destroy_registry=True)


@pytest.fixture()
def dbos_flask(
    config: DBOSConfig, cleanup_test_databases: None
) -> Generator[Tuple[DBOS, Flask], Any, None]:
    DBOS.destroy(destroy_registry=True)
    app = Flask(__name__)

    dbos = DBOS(flask=app, config=config)

    # This is for test convenience.
    #    Usually fastapi itself does launch, but we are not completing the fastapi lifecycle
    DBOS.launch()

    yield dbos, app
    DBOS.destroy(destroy_registry=True)


# Pretty-print test names
def pytest_collection_modifyitems(session: Any, config: Any, items: Any) -> None:
    for item in items:
        item._nodeid = "\n" + item.nodeid + "\n"


def queue_entries_are_cleaned_up(dbos: DBOS) -> bool:
    max_tries = 10
    success = False
    for i in range(max_tries):
        with dbos._sys_db.engine.begin() as c:
            query = (
                sa.select(sa.func.count())
                .select_from(SystemSchema.workflow_status)
                .where(
                    sa.and_(
                        SystemSchema.workflow_status.c.queue_name.isnot(None),
                        SystemSchema.workflow_status.c.status.in_(
                            ["ENQUEUED", "PENDING"]
                        ),
                    )
                )
            )
            row = c.execute(query).fetchone()
            assert row is not None
            count = row[0]
            if count == 0:
                success = True
                break
        time.sleep(1)
    return success


# Force exit after test success or failure with appropriate error code
_EXIT_CODE = None


def pytest_sessionfinish(session: Any, exitstatus: int) -> None:
    global _EXIT_CODE
    _EXIT_CODE = 0 if exitstatus == 0 else 1


def pytest_unconfigure(config: Any) -> None:
    code = _EXIT_CODE if _EXIT_CODE is not None else 1
    sys.exit(code)



import time
from typing import Any, Dict, Optional, cast

import psycopg
import sqlalchemy as sa
from sqlalchemy.exc import DBAPIError

from dbos._migration import ensure_dbos_schema, run_dbos_migrations

from ._logger import dbos_logger
from ._sys_db import SystemDatabase


class PostgresSystemDatabase(SystemDatabase):
    """PostgreSQL-specific implementation of SystemDatabase."""

    notification_conn: Optional[sa.PoolProxiedConnection] = None

    def _create_engine(
        self, system_database_url: str, engine_kwargs: Dict[str, Any]
    ) -> sa.Engine:
        url = sa.make_url(system_database_url).set(drivername="postgresql+psycopg")
        return sa.create_engine(url, **engine_kwargs)

    def run_migrations(self) -> None:
        """Run PostgreSQL-specific migrations."""
        if self._debug_mode:
            dbos_logger.warning("System database migrations are skipped in debug mode.")
            return
        system_db_url = self.engine.url
        sysdb_name = system_db_url.database
        # Unless we were provided an engine, if the system database does not already exist, create it
        if self.created_engine:
            engine = sa.create_engine(
                system_db_url.set(database="postgres"), **self._engine_kwargs
            )
            with engine.connect() as conn:
                conn.execution_options(isolation_level="AUTOCOMMIT")
                if not conn.execute(
                    sa.text("SELECT 1 FROM pg_database WHERE datname=:db_name"),
                    parameters={"db_name": sysdb_name},
                ).scalar():
                    dbos_logger.info(f"Creating system database {sysdb_name}")
                    conn.execute(sa.text(f"CREATE DATABASE {sysdb_name}"))
            engine.dispose()
        else:
            # If we were provided an engine, validate it can connect
            with self.engine.connect() as conn:
                conn.execute(sa.text("SELECT 1"))

        assert self.schema
        ensure_dbos_schema(self.engine, self.schema)
        run_dbos_migrations(self.engine, self.schema)

    def _cleanup_connections(self) -> None:
        """Clean up PostgreSQL-specific connections."""
        with self._listener_thread_lock:
            if self.notification_conn and self.notification_conn.dbapi_connection:
                self.notification_conn.dbapi_connection.close()
                self.notification_conn.invalidate()

    def _is_unique_constraint_violation(self, dbapi_error: DBAPIError) -> bool:
        """Check if the error is a unique constraint violation in PostgreSQL."""
        return dbapi_error.orig.sqlstate == "23505"  # type: ignore

    def _is_foreign_key_violation(self, dbapi_error: DBAPIError) -> bool:
        """Check if the error is a foreign key violation in PostgreSQL."""
        return dbapi_error.orig.sqlstate == "23503"  # type: ignore

    @staticmethod
    def _reset_system_database(database_url: str) -> None:
        """Reset the PostgreSQL system database by dropping it."""
        system_db_url = sa.make_url(database_url)
        sysdb_name = system_db_url.database

        if sysdb_name is None:
            raise ValueError(f"System database name not found in URL {system_db_url}")

        try:
            # Connect to postgres default database
            engine = sa.create_engine(
                system_db_url.set(database="postgres", drivername="postgresql+psycopg"),
                connect_args={"connect_timeout": 10},
            )

            with engine.connect() as conn:
                # Set autocommit required for database dropping
                conn.execution_options(isolation_level="AUTOCOMMIT")

                # Drop the database
                conn.execute(
                    sa.text(f"DROP DATABASE IF EXISTS {sysdb_name} WITH (FORCE)")
                )
            engine.dispose()
        except Exception as e:
            dbos_logger.error(f"Error resetting PostgreSQL system database: {str(e)}")
            raise e

    def _notification_listener(self) -> None:
        """Listen for PostgreSQL notifications using psycopg."""
        while self._run_background_processes:
            try:
                with self._listener_thread_lock:
                    self.notification_conn = self.engine.raw_connection()
                    self.notification_conn.detach()
                    psycopg_conn = cast(
                        psycopg.connection.Connection, self.notification_conn
                    )
                    psycopg_conn.set_autocommit(True)

                    psycopg_conn.execute("LISTEN dbos_notifications_channel")
                    psycopg_conn.execute("LISTEN dbos_workflow_events_channel")
                while self._run_background_processes:
                    gen = psycopg_conn.notifies()
                    for notify in gen:
                        channel = notify.channel
                        dbos_logger.debug(
                            f"Received notification on channel: {channel}, payload: {notify.payload}"
                        )
                        if channel == "dbos_notifications_channel":
                            if notify.payload:
                                condition = self.notifications_map.get(notify.payload)
                                if condition is None:
                                    # No condition found for this payload
                                    continue
                                condition.acquire()
                                condition.notify_all()
                                condition.release()
                                dbos_logger.debug(
                                    f"Signaled notifications condition for {notify.payload}"
                                )
                        elif channel == "dbos_workflow_events_channel":
                            if notify.payload:
                                condition = self.workflow_events_map.get(notify.payload)
                                if condition is None:
                                    # No condition found for this payload
                                    continue
                                condition.acquire()
                                condition.notify_all()
                                condition.release()
                                dbos_logger.debug(
                                    f"Signaled workflow_events condition for {notify.payload}"
                                )
                        else:
                            dbos_logger.error(f"Unknown channel: {channel}")
            except Exception as e:
                if self._run_background_processes:
                    dbos_logger.warning(f"Notification listener error: {e}")
                    time.sleep(1)
                    # Then the loop will try to reconnect and restart the listener
            finally:
                self._cleanup_connections()



# type: ignore

import os
from unittest.mock import mock_open
from urllib.parse import quote

import pytest
import pytest_mock
from sqlalchemy import event
from sqlalchemy.exc import OperationalError

# Public API
from dbos import DBOS, DBOSClient
from dbos._dbos_config import (
    ConfigFile,
    DBOSConfig,
    configure_db_engine_parameters,
    load_config,
    overwrite_config,
    process_config,
    translate_dbos_config_to_config_file,
)
from dbos._error import DBOSInitializationError
from dbos._schemas.system_database import SystemSchema

mock_filename = "dbos-config.yaml"
original_open = __builtins__["open"]


def generate_mock_open(filenames, mock_files):
    if not isinstance(filenames, list):
        filenames = [filenames]
    if not isinstance(mock_files, list):
        mock_files = [mock_files]

    def conditional_mock_open(*args, **kwargs):
        for filename, mock_file in zip(filenames, mock_files):
            if args[0] == filename:
                m = mock_open(read_data=mock_file)
                return m()
        return original_open(*args, **kwargs)

    return conditional_mock_open


"""
Test all the possible ways to configure DBOS.
- First test the "switches" in the DBOS.__init__() method, ensuring they find the config from the right place
- Then test each of the individual functions that process the config (load_config, overwrite_config, translate_dbos_config_to_config_file, process_config)
"""


####################
# DBOS launch
####################
def test_no_config_provided():
    with pytest.raises(TypeError) as exc_info:
        DBOS()
    assert "missing 1 required keyword-only argument: 'config'" in str(exc_info.value)


def test_dbosconfig_type_provided():
    config: DBOSConfig = {
        "name": "some-app",
        "database_url": f"postgres://postgres:{os.environ.get('PGPASSWORD', 'dbos')}@localhost:5432/some_app",
    }
    dbos = DBOS(config=config)
    assert dbos._config["name"] == "some-app"
    assert dbos._config["database_url"] == config["database_url"]
    dbos.destroy()


####################
# LOAD CONFIG
####################


def test_load_valid_config_file(mocker):
    mock_config = """
        name: "some-app"
        runtimeConfig:
            start:
                - "python3 main.py"
            admin_port: 8001
        database_url: "postgres://user:dbos@localhost:5432/dbname?connect_timeout=10&sslmode=require&sslrootcert=ca.pem"
        system_database_url: "postgres://user:dbos@localhost:5432/dbname_dbos_sys?connect_timeout=10&sslmode=require&sslrootcert=ca.pem"
        telemetry:
            OTLPExporter:
                logsEndpoint: 'fooLogs'
                tracesEndpoint: 'fooTraces'
    """
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open(mock_filename, mock_config)
    )

    configFile = load_config(mock_filename)
    assert configFile["name"] == "some-app"
    assert (
        configFile["database_url"]
        == f"postgres://user:dbos@localhost:5432/dbname?connect_timeout=10&sslmode=require&sslrootcert=ca.pem"
    )
    assert (
        configFile["system_database_url"]
        == f"postgres://user:dbos@localhost:5432/dbname_dbos_sys?connect_timeout=10&sslmode=require&sslrootcert=ca.pem"
    )

    assert configFile["telemetry"]["OTLPExporter"]["logsEndpoint"] == ["fooLogs"]
    assert configFile["telemetry"]["OTLPExporter"]["tracesEndpoint"] == ["fooTraces"]


def test_load_config_with_unset_database_url_env_var(mocker):
    mock_config = """
    name: "some-app"
    database_url: ${UNSET}
    """

    mocker.patch(
        "builtins.open", side_effect=generate_mock_open(mock_filename, mock_config)
    )

    configFile = load_config(mock_filename)
    assert configFile["name"] == "some-app"


def test_load_config_file_open_error(mocker):
    """Test handling when the config file can't be opened."""
    mocker.patch("builtins.open", side_effect=FileNotFoundError("File not found"))

    with pytest.raises(FileNotFoundError):
        load_config()


def test_load_config_file_not_a_dict(mocker):
    """Test handling when YAML doesn't parse to a dictionary."""
    mock_config = "just a string"
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open("dbos-config.yaml", mock_config)
    )

    with pytest.raises(DBOSInitializationError) as exc_info:
        load_config()

    assert "must contain a dictionary" in str(exc_info.value)


def test_load_config_file_custom_path():
    """Test parsing a config file from a custom path."""
    mock_config = """
    name: "test-app"
    """
    custom_path = "/custom/path/dbos-config.yaml"
    from unittest.mock import mock_open, patch

    with patch("builtins.open", mock_open(read_data=mock_config)) as mock_file:
        result = load_config(custom_path)
        mock_file.assert_called_with(custom_path, "r")
        assert result["name"] == "test-app"


####################
# PROCESS CONFIG
####################


# Full config provided
def test_process_config_full():
    config: ConfigFile = {
        "name": "some-app",
        "database_url": "postgres://user:password@localhost:7777/dbn?connect_timeout=1&sslmode=require&sslrootcert=ca.pem",
        "database": {
            "sys_db_pool_size": 27,
            "db_engine_kwargs": {"key": "value"},
            "migrate": ["alembic upgrade head"],
        },
        "runtimeConfig": {
            "start": ["python3 main.py"],
            "admin_port": 8001,
            "run_admin_server": False,
            "setup": ["echo 'hello'"],
        },
        "telemetry": {
            "logs": {
                "logLevel": "DEBUG",
            },
            "OTLPExporter": {
                "logsEndpoint": ["thelogsendpoint"],
                "tracesEndpoint": ["thetracesendpoint"],
            },
        },
        "env": {
            "FOO": "BAR",
        },
    }

    configFile = process_config(data=config)
    assert configFile["name"] == "some-app"
    assert (
        configFile["database_url"]
        == "postgres://user:password@localhost:7777/dbn?connect_timeout=1&sslmode=require&sslrootcert=ca.pem"
    )
    assert (
        configFile["system_database_url"]
        == f"postgres://user:password@localhost:7777/dbn_dbos_sys?connect_timeout=1&sslmode=require&sslrootcert=ca.pem"
    )
    assert configFile["database"]["migrate"] == ["alembic upgrade head"]
    assert configFile["database"]["db_engine_kwargs"] == {
        "key": "value",
        "pool_timeout": 30,
        "max_overflow": 0,
        "pool_size": 20,
        "pool_pre_ping": True,
        "connect_args": {"connect_timeout": 1},
    }
    assert configFile["database"]["sys_db_engine_kwargs"] == {
        "key": "value",
        "pool_timeout": 30,
        "max_overflow": 0,
        "pool_size": 27,
        "pool_pre_ping": True,
        "connect_args": {"connect_timeout": 1},
    }
    assert configFile["runtimeConfig"]["start"] == ["python3 main.py"]
    assert configFile["runtimeConfig"]["admin_port"] == 8001
    assert configFile["runtimeConfig"]["run_admin_server"] == False
    assert configFile["runtimeConfig"]["setup"] == ["echo 'hello'"]
    assert configFile["telemetry"]["logs"]["logLevel"] == "DEBUG"
    assert configFile["telemetry"]["OTLPExporter"]["logsEndpoint"] == [
        "thelogsendpoint"
    ]
    assert configFile["telemetry"]["OTLPExporter"]["tracesEndpoint"] == [
        "thetracesendpoint"
    ]
    assert configFile["env"]["FOO"] == "BAR"


def test_process_config_system_database():
    config: ConfigFile = {
        "name": "some-app",
        "database_url": "postgres://user:password@localhost:7777/dbn?connect_timeout=1&sslmode=require&sslrootcert=ca.pem",
        "system_database_url": "postgres://user:password@localhost:7778/dbn_sys?connect_timeout=1&sslmode=require&sslrootcert=ca.pem",
        "database": {
            "sys_db_pool_size": 27,
            "db_engine_kwargs": {"key": "value"},
            "migrate": ["alembic upgrade head"],
        },
    }

    configFile = process_config(data=config)
    assert configFile["name"] == "some-app"
    assert configFile["database_url"] == config["database_url"]
    assert configFile["system_database_url"] == config["system_database_url"]
    assert configFile["database"]["db_engine_kwargs"] == {
        "key": "value",
        "pool_timeout": 30,
        "max_overflow": 0,
        "pool_size": 20,
        "pool_pre_ping": True,
        "connect_args": {"connect_timeout": 1},
    }
    assert configFile["database"]["sys_db_engine_kwargs"] == {
        "key": "value",
        "pool_timeout": 30,
        "max_overflow": 0,
        "pool_size": 27,
        "pool_pre_ping": True,
        "connect_args": {"connect_timeout": 1},
    }


def test_process_config_only_system_database():
    config: ConfigFile = {
        "name": "some-app",
        "system_database_url": "postgres://user:password@localhost:7778/dbn_sys?connect_timeout=1&sslmode=require&sslrootcert=ca.pem",
    }

    configFile = process_config(data=config)
    assert configFile["name"] == "some-app"
    assert configFile["system_database_url"] == config["system_database_url"]
    assert configFile["database_url"] is None


def test_process_config_sqlite():
    config: ConfigFile = {
        "name": "some-app",
        "database_url": "sqlite:///test.sqlite",
        "system_database_url": "sqlite:///test.sys.sqlite",
    }

    configFile = process_config(data=config)
    assert configFile["name"] == "some-app"
    assert configFile["database_url"] == config["database_url"]
    assert configFile["system_database_url"] == config["system_database_url"]


def test_debug_override_database_url(mocker: pytest_mock.MockFixture):
    mocker.patch.dict(
        os.environ,
        {
            "DBOS_DBHOST": "fakehost",
            "DBOS_DBPORT": "1234",
            "DBOS_DBUSER": "fakeuser",
            "DBOS_DBPASSWORD": "fakepassword",
        },
    )
    config: ConfigFile = {
        "name": "some-app",
        "database_url": "postgres://user:password@localhost:7777/dbn?connect_timeout=1&sslmode=require&sslrootcert=ca.pem",
    }
    processed_config = process_config(data=config)
    assert (
        processed_config["database_url"]
        == "postgres://fakeuser:fakepassword@fakehost:1234/dbn?connect_timeout=1&sslmode=require&sslrootcert=ca.pem"
    )
    assert processed_config["name"] == "some-app"
    assert (
        processed_config["system_database_url"]
        == f"postgres://fakeuser:fakepassword@fakehost:1234/dbn{SystemSchema.sysdb_suffix}?connect_timeout=1&sslmode=require&sslrootcert=ca.pem"
    )
    assert processed_config["database"]["db_engine_kwargs"] is not None
    assert processed_config["database"]["sys_db_engine_kwargs"] is not None
    assert processed_config["runtimeConfig"]["run_admin_server"] == True
    assert processed_config["telemetry"]["logs"]["logLevel"] == "INFO"


def test_process_config_load_defaults():
    config: ConfigFile = {
        "name": "some-app",
    }
    processed_config = process_config(data=config)
    assert processed_config["name"] == "some-app"
    assert processed_config["database_url"] is None
    assert processed_config["system_database_url"] == f"sqlite:///some_app.sqlite"
    assert processed_config["database"]["db_engine_kwargs"] is not None
    assert processed_config["database"]["sys_db_engine_kwargs"] is not None
    assert processed_config["telemetry"]["logs"]["logLevel"] == "INFO"
    assert processed_config["runtimeConfig"]["run_admin_server"] == True


def test_process_config_load_default_with_None_database_url():
    config: ConfigFile = {
        "name": "some-app",
        "database_url": None,
    }
    processed_config = process_config(data=config)
    assert processed_config["name"] == "some-app"
    assert processed_config["database_url"] is None
    assert processed_config["system_database_url"] == f"sqlite:///some_app.sqlite"
    assert processed_config["database"]["db_engine_kwargs"] is not None
    assert processed_config["database"]["sys_db_engine_kwargs"] is not None
    assert processed_config["telemetry"]["logs"]["logLevel"] == "INFO"
    assert processed_config["runtimeConfig"]["run_admin_server"] == True


def test_process_config_load_default_with_empty_database_url():
    config: ConfigFile = {
        "name": "some-app",
        "database_url": "",
    }
    processed_config = process_config(data=config)
    assert processed_config["name"] == "some-app"
    assert processed_config["database_url"] is None
    assert processed_config["system_database_url"] == f"sqlite:///some_app.sqlite"
    assert processed_config["database"]["db_engine_kwargs"] is not None
    assert processed_config["database"]["sys_db_engine_kwargs"] is not None
    assert processed_config["telemetry"]["logs"]["logLevel"] == "INFO"
    assert processed_config["runtimeConfig"]["run_admin_server"] == True


def test_config_missing_name():
    config = {}
    with pytest.raises(DBOSInitializationError) as exc_info:
        process_config(data=config)

    assert "must specify an application name" in str(exc_info.value)


def test_config_bad_name():
    config: ConfigFile = {
        "name": "some app",
    }
    with pytest.raises(DBOSInitializationError) as exc_info:
        process_config(data=config)
    assert "Invalid app name" in str(exc_info.value)


####################
# PROCESS DB ENGINE KWARGS
####################
def test_configure_db_engine_parameters_defaults():
    """Test that default values are set when no pool sizes or kwargs are provided."""
    data: DatabaseConfig = {}

    configure_db_engine_parameters(data)

    assert data["db_engine_kwargs"] == {
        "pool_timeout": 30,
        "max_overflow": 0,
        "pool_size": 20,
        "pool_pre_ping": True,
        "connect_args": {"connect_timeout": 10},
    }
    assert data["sys_db_engine_kwargs"] == {
        "pool_timeout": 30,
        "max_overflow": 0,
        "pool_size": 20,
        "pool_pre_ping": True,
        "connect_args": {"connect_timeout": 10},
    }


def test_configure_db_engine_parameters_custom_sys_db_pool_sizes():
    """Test that custom pool sizes are preserved and used in engine kwargs."""
    data: DatabaseConfig = {"sys_db_pool_size": 35}

    configure_db_engine_parameters(data)

    assert data["db_engine_kwargs"] == {
        "pool_timeout": 30,
        "max_overflow": 0,
        "pool_size": 20,
        "pool_pre_ping": True,
        "connect_args": {"connect_timeout": 10},
    }
    assert data["sys_db_engine_kwargs"] == {
        "pool_timeout": 30,
        "max_overflow": 0,
        "pool_size": 35,
        "pool_pre_ping": True,
        "connect_args": {"connect_timeout": 10},
    }


def test_configure_db_engine_parameters_user_kwargs_override():
    """Test that user-provided db_engine_kwargs override defaults."""
    data: DatabaseConfig = {
        "sys_db_pool_size": 35,
        "db_engine_kwargs": {
            "pool_timeout": 60,
            "max_overflow": 10,
            "pool_pre_ping": True,
            "custom_param": "value",
            "pool_size": 50,
            "connect_args": {"connect_timeout": 30, "key": "value"},
        },
    }

    configure_db_engine_parameters(data)

    # User kwargs should override defaults and include custom params
    assert data["db_engine_kwargs"] == {
        "pool_timeout": 60,
        "max_overflow": 10,
        "pool_pre_ping": True,
        "custom_param": "value",
        "pool_size": 50,
        "connect_args": {"connect_timeout": 30, "key": "value"},
    }

    # System engine kwargs should use system pool size but same user overrides
    assert data["sys_db_engine_kwargs"] == {
        "pool_timeout": 60,
        "max_overflow": 10,
        "pool_pre_ping": True,
        "custom_param": "value",
        "pool_size": 35,
        "connect_args": {"connect_timeout": 30, "key": "value"},
    }


def test_configure_db_engine_parameters_user_kwargs_and_db_url_connect_timeout():
    """Test that user-provided db_engine_kwargs override defaults and include connect_timeout from db_url."""
    data: DatabaseConfig = {
        "db_engine_kwargs": {
            "pool_timeout": 60,
            "pool_pre_ping": True,
            "custom_param": "value",
            "pool_size": 50,
        },
    }

    configure_db_engine_parameters(data, connect_timeout=22)

    # User kwargs should override defaults and include custom params
    assert data["db_engine_kwargs"] == {
        "pool_timeout": 60,
        "max_overflow": 0,
        "pool_pre_ping": True,
        "custom_param": "value",
        "pool_size": 50,
        "connect_args": {"connect_timeout": 22},
    }

    # System engine kwargs should use system pool size but same user overrides
    assert data["sys_db_engine_kwargs"] == {
        "pool_timeout": 60,
        "max_overflow": 0,
        "pool_pre_ping": True,
        "custom_param": "value",
        "pool_size": 50,
        "connect_args": {"connect_timeout": 22},
    }


def test_configure_db_engine_parameters_user_kwargs_plus_db_url_connect_timeout():
    """Test that user-provided db_engine_kwargs override defaults and connect_timeout from db_url."""
    data: DatabaseConfig = {
        "db_engine_kwargs": {
            "pool_timeout": 60,
            "pool_pre_ping": True,
            "custom_param": "value",
            "pool_size": 50,
            "connect_args": {"connect_timeout": 1},
        },
    }

    configure_db_engine_parameters(data, connect_timeout=22)

    # User kwargs should override defaults and include custom params
    assert data["db_engine_kwargs"] == {
        "pool_timeout": 60,
        "max_overflow": 0,
        "pool_pre_ping": True,
        "custom_param": "value",
        "pool_size": 50,
        "connect_args": {"connect_timeout": 1},
    }

    # System engine kwargs should use system pool size but same user overrides
    assert data["sys_db_engine_kwargs"] == {
        "pool_timeout": 60,
        "max_overflow": 0,
        "pool_pre_ping": True,
        "custom_param": "value",
        "pool_size": 50,
        "connect_args": {"connect_timeout": 1},
    }


def test_configure_db_engine_parameters_user_kwargs_mixed_params():
    """Test that user-provided db_engine_kwargs override defaults."""
    data: DatabaseConfig = {
        "db_engine_kwargs": {
            "pool_timeout": 60,
            "pool_pre_ping": True,
            "custom_param": "value",
            "pool_size": 50,
        }
    }

    configure_db_engine_parameters(data)

    # User kwargs should override defaults and include custom params
    assert data["db_engine_kwargs"] == {
        "pool_timeout": 60,
        "max_overflow": 0,
        "pool_pre_ping": True,
        "custom_param": "value",
        "pool_size": 50,
        "connect_args": {"connect_timeout": 10},
    }

    # System engine kwargs should use system pool size but same user overrides
    assert data["sys_db_engine_kwargs"] == {
        "pool_timeout": 60,
        "max_overflow": 0,
        "pool_pre_ping": True,
        "custom_param": "value",
        "pool_size": 50,
        "connect_args": {"connect_timeout": 10},
    }


def test_configure_db_engine_parameters_empty_user_kwargs():
    """Test handling of empty user kwargs dict."""
    data: DatabaseConfig = {"db_engine_kwargs": {}}

    configure_db_engine_parameters(data)

    assert data["db_engine_kwargs"] == {
        "pool_timeout": 30,
        "max_overflow": 0,
        "pool_size": 20,
        "pool_pre_ping": True,
        "connect_args": {"connect_timeout": 10},
    }
    assert data["sys_db_engine_kwargs"] == {
        "pool_timeout": 30,
        "max_overflow": 0,
        "pool_size": 20,
        "pool_pre_ping": True,
        "connect_args": {"connect_timeout": 10},
    }


####################
# VALIDATE DATABASE URL
####################


def test_process_config_with_wrong_db_url():
    # Missing username
    config: ConfigFile = {
        "name": "some-app",
        "database_url": "postgres://:password@h:1234/dbname",
    }
    with pytest.raises(DBOSInitializationError) as exc_info:
        process_config(data=config)
    assert "Username must be specified in the connection URL" in str(exc_info.value)

    # Missing host
    config: ConfigFile = {
        "name": "some-app",
        "database_url": "postgres://user:password@:1234/dbname",
    }
    with pytest.raises(DBOSInitializationError) as exc_info:
        process_config(data=config)
    assert "Host must be specified in the connection URL" in str(exc_info.value)

    # Missing dbname
    config: ConfigFile = {
        "name": "some-app",
        "database_url": "postgres://user:password@h:1234",
    }
    with pytest.raises(DBOSInitializationError) as exc_info:
        process_config(data=config)
    assert "Database name must be specified in the connection URL" in str(
        exc_info.value
    )

    # Missing dbname in system database
    config: ConfigFile = {
        "name": "some-app",
        "system_database_url": "postgres://user:password@h:1234",
    }
    with pytest.raises(DBOSInitializationError) as exc_info:
        process_config(data=config)
    assert "Database name must be specified in the connection URL" in str(
        exc_info.value
    )


def test_database_url_no_password(skip_with_sqlite: None):
    """Test that the database URL can be provided without a password."""
    expected_url = "postgresql://postgres@localhost:5432/dbostestpy?sslmode=disable"
    config: DBOSConfig = {
        "name": "some-app",
        "database_url": expected_url,
    }
    processed_config = translate_dbos_config_to_config_file(config)
    assert processed_config["name"] == "some-app"
    assert processed_config["database_url"] == expected_url


####################
# TRANSLATE DBOSConfig to ConfigFile
####################


def test_translate_dbosconfig_full_input():
    # Give all fields
    config: DBOSConfig = {
        "name": "test-app",
        "database_url": "postgres://user:password@localhost:5432/dbname?connect_timeout=11&sslmode=require&sslrootcert=ca.pem",
        "sys_db_pool_size": 27,
        "db_engine_kwargs": {"key": "value"},
        "log_level": "DEBUG",
        "otlp_traces_endpoints": ["http://otel:7777", "notused"],
        "admin_port": 8001,
        "run_admin_server": False,
        "dbos_system_schema": "foobar",
    }

    translated_config = translate_dbos_config_to_config_file(config)

    assert translated_config["name"] == "test-app"
    assert translated_config["database_url"] == config["database_url"]
    assert translated_config["database"]["sys_db_pool_size"] == 27
    assert translated_config["database"]["db_engine_kwargs"] == {"key": "value"}
    assert translated_config["telemetry"]["logs"]["logLevel"] == "DEBUG"
    assert translated_config["telemetry"]["OTLPExporter"]["tracesEndpoint"] == [
        "http://otel:7777",
        "notused",
    ]
    assert translated_config["telemetry"]["OTLPExporter"]["logsEndpoint"] == []
    assert translated_config["telemetry"]["disable_otlp"] == True
    assert translated_config["runtimeConfig"]["admin_port"] == 8001
    assert translated_config["runtimeConfig"]["run_admin_server"] == False
    assert translated_config["dbos_system_schema"] == "foobar"
    assert "start" not in translated_config["runtimeConfig"]
    assert "setup" not in translated_config["runtimeConfig"]
    assert "env" not in translated_config


def test_translate_dbosconfig_minimal_input():
    config: DBOSConfig = {
        "name": "test-app",
    }
    translated_config = translate_dbos_config_to_config_file(config)

    assert translated_config["name"] == "test-app"
    assert translated_config["telemetry"]["logs"]["logLevel"] == "INFO"
    assert translated_config["runtimeConfig"]["run_admin_server"] == True
    assert "admin_port" not in translated_config["runtimeConfig"]
    assert "database" not in translated_config
    assert "env" not in translated_config


def test_translate_dbosconfig_just_sys_db_pool_size():
    config: DBOSConfig = {
        "name": "test-app",
        "sys_db_pool_size": 27,
    }
    translated_config = translate_dbos_config_to_config_file(config)

    assert translated_config["database"]["sys_db_pool_size"] == 27
    assert "env" not in translated_config


def test_translate_dbosconfig_just_db_engine_kwargs():
    config: DBOSConfig = {
        "name": "test-app",
        "db_engine_kwargs": {"key": "value"},
    }
    translated_config = translate_dbos_config_to_config_file(config)

    assert translated_config["database"]["db_engine_kwargs"] == {"key": "value"}
    assert "sys_db_pool_size" not in translated_config["database"]
    assert "env" not in translated_config
    assert "admin_port" not in translated_config["runtimeConfig"]


def test_translate_dbosconfig_just_admin_port():
    config: DBOSConfig = {
        "name": "test-app",
        "admin_port": 8001,
    }
    translated_config = translate_dbos_config_to_config_file(config)

    assert translated_config["runtimeConfig"]["admin_port"] == 8001
    assert translated_config["runtimeConfig"]["run_admin_server"] == True
    assert "env" not in translated_config
    assert "database" not in translated_config


def test_translate_dbosconfig_just_run_admin_server():
    config: DBOSConfig = {
        "name": "test-app",
        "run_admin_server": True,
    }
    translated_config = translate_dbos_config_to_config_file(config)

    assert translated_config["name"] == "test-app"
    assert translated_config["telemetry"]["logs"]["logLevel"] == "INFO"
    assert translated_config["runtimeConfig"]["run_admin_server"] == True
    assert "admin_port" not in translated_config["runtimeConfig"]
    assert "env" not in translated_config
    assert "database" not in translated_config


def test_translate_empty_otlp_traces_endpoints():
    # Give an empty OTLP traces endpoint list
    config: DBOSConfig = {
        "name": "test-app",
        "otlp_traces_endpoints": [],
    }
    translated_config = translate_dbos_config_to_config_file(config)
    assert len(translated_config["telemetry"]["OTLPExporter"]["logsEndpoint"]) == 0
    assert len(translated_config["telemetry"]["OTLPExporter"]["tracesEndpoint"]) == 0
    assert translated_config["telemetry"]["logs"]["logLevel"] == "INFO"


def test_translate_ignores_otlp_traces_not_list():
    # Give an empty OTLP traces endpoint list
    config: DBOSConfig = {
        "name": "test-app",
        "otlp_traces_endpoints": "http://otel:7777",
    }
    translated_config = translate_dbos_config_to_config_file(config)
    assert translated_config["name"] == "test-app"
    assert len(translated_config["telemetry"]["OTLPExporter"]["logsEndpoint"]) == 0
    assert len(translated_config["telemetry"]["OTLPExporter"]["tracesEndpoint"]) == 0


def test_translate_missing_name():
    with pytest.raises(DBOSInitializationError) as exc_info:
        translate_dbos_config_to_config_file({})
    assert (
        "Error initializing DBOS Transact: Configuration must specify an application name"
        in str(exc_info.value)
    )


def test_translate_application_database_url():
    config: DBOSConfig = {
        "name": "test-app",
        "application_database_url": "postgres://user:password@localhost:5432/dbname?connect_timeout=11&sslmode=require&sslrootcert=ca.pem",
        "system_database_url": "postgres://user:password@localhost:5432/dbname_sys?connect_timeout=11&sslmode=require&sslrootcert=ca.pem",
    }
    translated_config = translate_dbos_config_to_config_file(config)
    assert translated_config["name"] == "test-app"
    assert translated_config["database_url"] == config["application_database_url"]
    assert translated_config["system_database_url"] == config["system_database_url"]


####################
# CONFIG OVERWRITE
####################


def test_overwrite_config(mocker):
    # Setup a typical dbos-config.yaml file
    mock_config = """
    name: "stock-prices"
    language: "python"
    database:
        migrate:
            - alembic upgrade head
    telemetry:
        logs:
            logLevel: INFO
        OTLPExporter:
            logsEndpoint: thelogsendpoint
            tracesEndpoint:  thetracesendpoint
    runtimeConfig:
        start:
            - "a start command"
    """
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open("dbos-config.yaml", mock_config)
    )

    provided_config: ConfigFile = {
        "name": "test-app",
        "database": {},
        "telemetry": {
            "OTLPExporter": {
                "tracesEndpoint": ["a"],
                "logsEndpoint": ["b"],
            },
            "logs": {
                "logLevel": "DEBUG",
            },
        },
        "runtimeConfig": {
            "admin_port": 8001,
            "run_admin_server": True,
        },
        "env": {
            "FOO": "BAR",
        },
    }
    exported_db_url = "postgres://dbosadmin:pwd@hostname:1234/appdbname?connect_timeout=10000&sslmode=require&sslrootcert=cert.pem"
    os.environ["DBOS_DATABASE_URL"] = exported_db_url
    exported_sys_db_url = "postgres://dbosadmin:pwd@hostname:1234/appdbname_dbos_sys?connect_timeout=10000&sslmode=require&sslrootcert=cert.pem"
    os.environ["DBOS_SYSTEM_DATABASE_URL"] = exported_sys_db_url

    config = overwrite_config(provided_config)

    assert config["name"] == "stock-prices"
    assert config["database_url"] == exported_db_url
    assert config["system_database_url"] == exported_sys_db_url
    assert "sys_db_pool_size" not in config["database"]
    assert config["telemetry"]["logs"]["logLevel"] == "DEBUG"
    assert config["telemetry"]["OTLPExporter"]["tracesEndpoint"] == [
        "a",
        "thetracesendpoint",
    ]
    assert config["telemetry"]["OTLPExporter"]["logsEndpoint"] == [
        "b",
        "thelogsendpoint",
    ]
    assert config["telemetry"]["disable_otlp"] == False
    assert "admin_port" not in config["runtimeConfig"]
    assert "run_admin_server" not in config["runtimeConfig"]
    assert "env" not in config

    del os.environ["DBOS_DATABASE_URL"]
    del os.environ["DBOS_SYSTEM_DATABASE_URL"]


def test_overwrite_config_minimal(mocker):
    mock_config = """
    name: "stock-prices"
    language: "python"
    database:
        migrate:
            - alembic upgrade head
    telemetry:
        OTLPExporter:
            logsEndpoint: thelogsendpoint
            tracesEndpoint:  thetracesendpoint
    runtimeConfig:
        start:
            - "a start command"
        setup:
            - "echo 'hello'"
    """
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open("dbos-config.yaml", mock_config)
    )

    provided_config: ConfigFile = {
        "name": "test-app",
        "dbos_system_schema": "foobar",
    }

    exported_db_url = "postgres://dbosadmin:pwd@hostname:1234/appdbname?connect_timeout=10000&sslmode=require&sslrootcert=cert.pem"
    os.environ["DBOS_DATABASE_URL"] = exported_db_url
    exported_sys_db_url = "postgres://dbosadmin:pwd@hostname:1234/appdbname_dbos_sys?connect_timeout=10000&sslmode=require&sslrootcert=cert.pem"
    os.environ["DBOS_SYSTEM_DATABASE_URL"] = exported_sys_db_url

    config = overwrite_config(provided_config)

    assert config["name"] == "stock-prices"
    assert config["database_url"] == exported_db_url
    assert config["system_database_url"] == exported_sys_db_url
    assert config["dbos_system_schema"] == "dbos"
    assert config["telemetry"]["OTLPExporter"]["tracesEndpoint"] == [
        "thetracesendpoint"
    ]
    assert config["telemetry"]["OTLPExporter"]["logsEndpoint"] == ["thelogsendpoint"]
    assert "runtimeConfig" not in config
    assert "env" not in config

    del os.environ["DBOS_DATABASE_URL"]
    del os.environ["DBOS_SYSTEM_DATABASE_URL"]


def test_overwrite_config_has_telemetry(mocker):
    mock_config = """
    name: "stock-prices"
    language: "python"
    database:
        migrate:
            - alembic upgrade head
    telemetry:
        OTLPExporter:
            logsEndpoint: thelogsendpoint
            tracesEndpoint:  thetracesendpoint
    runtimeConfig:
        start:
            - "a start command"
        setup:
            - "echo 'hello'"
    """
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open("dbos-config.yaml", mock_config)
    )

    provided_config: ConfigFile = {
        "name": "test-app",
        "telemetry": {"logs": {"logLevel": "DEBUG"}},
    }

    exported_db_url = "postgres://dbosadmin:pwd@hostname:1234/appdbname?connect_timeout=10000&sslmode=require&sslrootcert=cert.pem"
    os.environ["DBOS_DATABASE_URL"] = exported_db_url
    exported_sys_db_url = "postgres://dbosadmin:pwd@hostname:1234/appdbname_dbos_sys?connect_timeout=10000&sslmode=require&sslrootcert=cert.pem"
    os.environ["DBOS_SYSTEM_DATABASE_URL"] = exported_sys_db_url

    config = overwrite_config(provided_config)

    assert config["name"] == "stock-prices"
    assert config["database_url"] == exported_db_url
    assert config["system_database_url"] == exported_sys_db_url
    assert config["telemetry"]["OTLPExporter"]["tracesEndpoint"] == [
        "thetracesendpoint"
    ]
    assert config["telemetry"]["OTLPExporter"]["logsEndpoint"] == ["thelogsendpoint"]
    assert config["telemetry"]["logs"]["logLevel"] == "DEBUG"
    assert "runtimeConfig" not in config
    assert "env" not in config

    del os.environ["DBOS_DATABASE_URL"]
    del os.environ["DBOS_SYSTEM_DATABASE_URL"]


# Not expected in practice, but exercise the code path
def test_overwrite_config_no_telemetry_in_file(mocker):
    mock_config = """
    name: "stock-prices"
    language: "python"
    """
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open("dbos-config.yaml", mock_config)
    )

    provided_config: ConfigFile = {
        "name": "test-app",
        "telemetry": {"logs": {"logLevel": "DEBUG"}},
    }

    exported_db_url = "postgres://dbosadmin:pwd@hostname:1234/appdbname?connect_timeout=10000&sslmode=require&sslrootcert=cert.pem"
    os.environ["DBOS_DATABASE_URL"] = exported_db_url
    exported_sys_db_url = "postgres://dbosadmin:pwd@hostname:1234/appdbname_dbos_sys?connect_timeout=10000&sslmode=require&sslrootcert=cert.pem"
    os.environ["DBOS_SYSTEM_DATABASE_URL"] = exported_sys_db_url

    config = overwrite_config(provided_config)
    # Test that telemetry from provided_config is preserved
    assert config["database_url"] == exported_db_url
    assert config["system_database_url"] == exported_sys_db_url
    assert config["telemetry"]["logs"]["logLevel"] == "DEBUG"
    assert config["telemetry"]["OTLPExporter"] == {
        "tracesEndpoint": [],
        "logsEndpoint": [],
    }

    del os.environ["DBOS_DATABASE_URL"]
    del os.environ["DBOS_SYSTEM_DATABASE_URL"]


# Not expected in practice, but exercise the code path
def test_overwrite_config_no_otlp_in_file(mocker):
    mock_config = """
    name: "stock-prices"
    language: "python"
    telemetry:
        logs:
            logLevel: INFO
    """
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open("dbos-config.yaml", mock_config)
    )

    provided_config: ConfigFile = {
        "name": "test-app",
        "telemetry": {
            "OTLPExporter": {
                "tracesEndpoint": ["original-trace"],
                "logsEndpoint": ["original-log"],
            }
        },
    }

    exported_db_url = "postgres://dbosadmin:pwd@hostname:1234/appdbname?connect_timeout=10000&sslmode=require&sslrootcert=cert.pem"
    os.environ["DBOS_DATABASE_URL"] = exported_db_url
    exported_sys_db_url = "postgres://dbosadmin:pwd@hostname:1234/appdbname_dbos_sys?connect_timeout=10000&sslmode=require&sslrootcert=cert.pem"
    os.environ["DBOS_SYSTEM_DATABASE_URL"] = exported_sys_db_url

    config = overwrite_config(provided_config)
    assert config["database_url"] == exported_db_url
    assert config["system_database_url"] == exported_sys_db_url
    # Test that OTLPExporter from provided_config is preserved
    assert config["telemetry"]["OTLPExporter"]["tracesEndpoint"] == ["original-trace"]
    assert config["telemetry"]["OTLPExporter"]["logsEndpoint"] == ["original-log"]
    assert "logs" not in config["telemetry"]

    del os.environ["DBOS_DATABASE_URL"]
    del os.environ["DBOS_SYSTEM_DATABASE_URL"]


def test_overwrite_config_with_provided_database_url(mocker):
    mock_config = """
    name: "stock-prices"
    language: "python"
    database:
        migrate:
            - alembic upgrade head
    telemetry:
        OTLPExporter:
            logsEndpoint: thelogsendpoint
            tracesEndpoint:  thetracesendpoint
    runtimeConfig:
        start:
            - "a start command"
        setup:
            - "echo 'hello'"
    """
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open("dbos-config.yaml", mock_config)
    )

    provided_config: ConfigFile = {
        "name": "test-app",
        "database_url": "ignored",
    }

    exported_db_url = "postgres://dbosadmin:pwd@hostname:1234/appdbname?connect_timeout=10000&sslmode=require&sslrootcert=cert.pem"
    os.environ["DBOS_DATABASE_URL"] = exported_db_url
    exported_sys_db_url = "postgres://dbosadmin:pwd@hostname:1234/appdbname_dbos_sys?connect_timeout=10000&sslmode=require&sslrootcert=cert.pem"
    os.environ["DBOS_SYSTEM_DATABASE_URL"] = exported_sys_db_url

    config = overwrite_config(provided_config)

    assert config["name"] == "stock-prices"
    assert config["database_url"] == exported_db_url
    assert config["system_database_url"] == exported_sys_db_url
    assert config["telemetry"]["OTLPExporter"]["tracesEndpoint"] == [
        "thetracesendpoint"
    ]
    assert config["telemetry"]["OTLPExporter"]["logsEndpoint"] == ["thelogsendpoint"]
    assert "runtimeConfig" not in config
    assert "env" not in config

    del os.environ["DBOS_DATABASE_URL"]
    del os.environ["DBOS_SYSTEM_DATABASE_URL"]


def test_overwrite_config_missing_dbos_database_url(mocker):
    mock_config = """
    name: "stock-prices"
    """

    mocker.patch(
        "builtins.open", side_effect=generate_mock_open("dbos-config.yaml", mock_config)
    )

    provided_config: ConfigFile = {
        "name": "test-app",
    }
    with pytest.raises(DBOSInitializationError) as exc_info:
        overwrite_config(provided_config)
    assert (
        "DBOS_DATABASE_URL environment variable is not set. This is required to connect to the database."
        in str(exc_info.value)
    )


####################
# DATABASES CONNECTION POOLS
####################


def test_configured_pool_default():
    DBOS.destroy()
    config: DBOSConfig = {
        "name": "test-app",
        "database_url": f"postgres://postgres:{quote(os.environ.get('PGPASSWORD', 'dbos'))}@localhost:5432/postgres",
    }

    dbos = DBOS(config=config)
    dbos.launch()
    assert dbos._app_db.engine.pool._pool.maxsize == 20
    assert dbos._app_db.engine.pool._timeout == 30
    assert dbos._app_db.engine.pool._max_overflow == 0
    assert dbos._app_db.engine.pool._pre_ping == True

    assert dbos._sys_db.engine.pool._pool.maxsize == 20
    assert dbos._sys_db.engine.pool._timeout == 30
    assert dbos._sys_db.engine.pool._max_overflow == 0
    assert dbos._sys_db.engine.pool._pre_ping == True

    # force the release of connections so we can intercept on connect.
    app_db_engine = dbos._app_db.engine
    app_db_engine.dispose()

    @event.listens_for(app_db_engine, "connect")
    def inspect_connection(dbapi_connection, connection_record):
        connect_timeout = dbapi_connection.info.get_parameters()["connect_timeout"]
        assert connect_timeout == "10"

    with app_db_engine.connect() as conn:
        pass

    dbos.destroy()


def test_configured_pool_custom_url():
    DBOS.destroy()
    config: DBOSConfig = {
        "name": "test-app",
        "database_url": f"postgres://postgres:{quote(os.environ.get('PGPASSWORD', 'dbos'))}@localhost:5432/postgres",
        "system_database_url": f"postgres://postgres:{quote(os.environ.get('PGPASSWORD', 'dbos'))}@localhost:5432/dbostesturl",
    }

    dbos = DBOS(config=config)
    dbos.launch()
    assert "postgres" in dbos._app_db.engine.url
    assert "dbostesturl" in dbos._sys_db.engine.url

    dbos.destroy()


def test_configured_pool_user_provided():
    DBOS.destroy()
    config: DBOSConfig = {
        "name": "test-app",
        "sys_db_pool_size": 43,
        "database_url": f"postgres://postgres:{quote(os.environ.get('PGPASSWORD', 'dbos'))}@localhost:5432/postgres",
        "db_engine_kwargs": {
            "pool_size": 22,
            "pool_timeout": 42,
            "max_overflow": 27,
            "pool_pre_ping": True,
            "connect_args": {"connect_timeout": 7},
        },
    }

    dbos = DBOS(config=config)
    dbos.launch()
    assert dbos._app_db.engine.pool._pool.maxsize == 22
    assert dbos._app_db.engine.pool._timeout == 42
    assert dbos._app_db.engine.pool._max_overflow == 27
    assert dbos._app_db.engine.pool._pre_ping == True

    assert dbos._sys_db.engine.pool._pool.maxsize == 43
    assert dbos._sys_db.engine.pool._timeout == 42
    assert dbos._sys_db.engine.pool._max_overflow == 27
    assert dbos._sys_db.engine.pool._pre_ping == True

    # force the release of connections so we can intercept on connect.
    app_db_engine = dbos._app_db.engine
    app_db_engine.dispose()

    @event.listens_for(app_db_engine, "connect")
    def inspect_connection(dbapi_connection, connection_record):
        connect_timeout = dbapi_connection.info.get_parameters()["connect_timeout"]
        assert connect_timeout == "7"

    with app_db_engine.connect() as conn:
        pass

    dbos.destroy()


def test_configured_pool_user_provided_dburl_connect_timeout():
    DBOS.destroy()
    config: DBOSConfig = {
        "name": "test-app",
        "sys_db_pool_size": 43,
        "database_url": f"postgres://postgres:{quote(os.environ.get('PGPASSWORD', 'dbos'))}@localhost:5432/postgres?connect_timeout=22",
        "db_engine_kwargs": {
            "pool_size": 22,
            "pool_timeout": 42,
            "max_overflow": 27,
            "pool_pre_ping": True,
        },
    }

    dbos = DBOS(config=config)
    dbos.launch()
    assert dbos._app_db.engine.pool._pool.maxsize == 22
    assert dbos._app_db.engine.pool._timeout == 42
    assert dbos._app_db.engine.pool._max_overflow == 27
    assert dbos._app_db.engine.pool._pre_ping == True

    assert dbos._sys_db.engine.pool._pool.maxsize == 43
    assert dbos._sys_db.engine.pool._timeout == 42
    assert dbos._sys_db.engine.pool._max_overflow == 27
    assert dbos._sys_db.engine.pool._pre_ping == True

    # force the release of connections so we can intercept on connect.
    app_db_engine = dbos._app_db.engine
    app_db_engine.dispose()

    @event.listens_for(app_db_engine, "connect")
    def inspect_connection(dbapi_connection, connection_record):
        connect_timeout = dbapi_connection.info.get_parameters()["connect_timeout"]
        assert connect_timeout == "22"

    with app_db_engine.connect() as conn:
        pass

    dbos.destroy()


def test_configured_pool_user_provided_dburl_connect_timeout_precedence():
    DBOS.destroy()
    config: DBOSConfig = {
        "name": "test-app",
        "sys_db_pool_size": 43,
        "database_url": f"postgres://postgres:{quote(os.environ.get('PGPASSWORD', 'dbos'))}@localhost:5432/postgres?connect_timeout=22",  # connect_args will take precedence
        "db_engine_kwargs": {
            "pool_size": 22,
            "pool_timeout": 42,
            "max_overflow": 27,
            "pool_pre_ping": True,
            "connect_args": {"connect_timeout": 7},
        },
    }

    dbos = DBOS(config=config)
    dbos.launch()
    assert dbos._app_db.engine.pool._pool.maxsize == 22
    assert dbos._app_db.engine.pool._timeout == 42
    assert dbos._app_db.engine.pool._max_overflow == 27
    assert dbos._app_db.engine.pool._pre_ping == True

    assert dbos._sys_db.engine.pool._pool.maxsize == 43
    assert dbos._sys_db.engine.pool._timeout == 42
    assert dbos._sys_db.engine.pool._max_overflow == 27
    assert dbos._sys_db.engine.pool._pre_ping == True

    # force the release of connections so we can intercept on connect.
    app_db_engine = dbos._app_db.engine
    app_db_engine.dispose()

    @event.listens_for(app_db_engine, "connect")
    def inspect_connection(dbapi_connection, connection_record):
        connect_timeout = dbapi_connection.info.get_parameters()["connect_timeout"]
        assert connect_timeout == "7"

    with app_db_engine.connect() as conn:
        pass

    dbos.destroy()


def test_pool_connection_times_out_by_default():
    import socket

    ipv4_addr = socket.gethostbyname("example.com")

    DBOS.destroy()
    config: DBOSConfig = {
        "name": "test-app",
        "database_url": f"postgres://postgres:dbos@{ipv4_addr}/postgres",
    }

    dbos = DBOS(config=config)
    with pytest.raises(OperationalError) as exc_info:
        dbos.launch()

    assert "timeout" in str(exc_info.value).lower()
    dbos.destroy()



import asyncio
import threading
from typing import Any, Coroutine, Optional, TypeVar


class BackgroundEventLoop:
    """
    This is the event loop to which DBOS submits any coroutines that are not started from within an event loop.
    In particular, coroutines submitted to queues (such as from scheduled workflows) run on this event loop.

    If a main event loop is known (whether because an event loop existed in the thread that called DBOS.launch
    or because a FastAPI event loop was detected) then coroutines are submitted there instead.
    """

    def __init__(self) -> None:
        self._main_loop: Optional[asyncio.AbstractEventLoop] = None
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._thread: Optional[threading.Thread] = None
        self._running = False
        self._ready = threading.Event()

    def start(self) -> None:
        if self._running:
            return

        self.set_main_loop()
        self._thread = threading.Thread(target=self._run_event_loop, daemon=True)
        self._thread.start()
        self._ready.wait()  # Wait until the loop is running

    def stop(self) -> None:
        if not self._running or self._loop is None or self._thread is None:
            return

        asyncio.run_coroutine_threadsafe(self._shutdown(), self._loop)
        self._thread.join()
        self._running = False

    def _run_event_loop(self) -> None:
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)

        self._running = True
        self._ready.set()  # Signal that the loop is ready

        try:
            self._loop.run_forever()
        finally:
            self._loop.close()

    async def _shutdown(self) -> None:
        if self._loop is None:
            raise RuntimeError("Event loop not started")
        tasks = [
            task
            for task in asyncio.all_tasks(self._loop)
            if task is not asyncio.current_task(self._loop)
        ]

        for task in tasks:
            task.cancel()

        await asyncio.gather(*tasks, return_exceptions=True)
        self._loop.stop()

    def set_main_loop(self) -> None:
        """
        Set the main loop to the currently running event loop.
        Should be called from the main thread.
        """
        try:
            self._main_loop = asyncio.get_running_loop()
        except:
            # There's no running event loop to set
            pass

    T = TypeVar("T")

    def submit_coroutine(self, coro: Coroutine[Any, Any, T]) -> T:
        """Submit a coroutine to the background event loop"""
        if self._main_loop is not None and self._main_loop.is_running():
            return asyncio.run_coroutine_threadsafe(coro, self._main_loop).result()
        if self._loop is None:
            raise RuntimeError("Event loop not started")
        return asyncio.run_coroutine_threadsafe(coro, self._loop).result()



import json
import os
import shutil
import signal
import subprocess
import tempfile
import time
import urllib.error
import urllib.request

import requests
import sqlalchemy as sa
import yaml

from dbos._dbos_config import DBOSConfig
from tests.conftest import using_sqlite


def test_package(
    build_wheel: str, db_engine: sa.Engine, skip_with_sqlite: None
) -> None:

    # Clean up the database from previous runs
    for template_name in ["dbos-db-starter", "dbos-app-starter"]:
        db_starter = template_name == "dbos-db-starter"
        app_db_name = template_name.replace("-", "_")
        sys_db_name = f"{app_db_name}_dbos_sys"
        with db_engine.connect() as connection:
            connection.execution_options(isolation_level="AUTOCOMMIT")
            connection.execute(sa.text(f"DROP DATABASE IF EXISTS {app_db_name}"))
            connection.execute(sa.text(f"DROP DATABASE IF EXISTS {sys_db_name}"))

        with tempfile.TemporaryDirectory() as temp_path:
            temp_path = tempfile.mkdtemp(prefix="dbos-")
            wheel_path = os.path.abspath(build_wheel)

            # Create a new virtual environment in the temp directory
            venv_path = os.path.join(temp_path, ".venv")
            if os.path.exists(venv_path):
                shutil.rmtree(venv_path)

            # To create a venv, we need the system Python executable. TODO: Don't hardcode the path.
            subprocess.check_call(
                [os.path.join("/", "usr", "bin", "python3"), "-m", "venv", venv_path]
            )

            venv = os.environ.copy()
            venv["PATH"] = f"{os.path.join(venv_path, 'bin')}:{venv['PATH']}"
            venv["VIRTUAL_ENV"] = venv_path
            venv["DBOS_DATABASE_URL"] = db_engine.url.set(
                database=app_db_name
            ).render_as_string(hide_password=False)
            venv["DBOS_SYSTEM_DATABASE_URL"] = db_engine.url.set(
                database=sys_db_name
            ).render_as_string(hide_password=False)

            # Install the dbos package into the virtual environment
            subprocess.check_call(
                ["pip", "install", wheel_path], cwd=temp_path, env=venv
            )

            # Next, verify a simple DBOS-only script runs
            subprocess.check_call(
                ["python3", "tests/script_without_fastapi.py"], env=venv
            )

            # Install FastAPI into the virtual environment
            subprocess.check_call(
                ["pip", "install", "fastapi[standard]"], cwd=temp_path, env=venv
            )

            # initalize the app with dbos scaffolding
            subprocess.check_call(
                ["dbos", "init", template_name, "--template", template_name],
                cwd=temp_path,
            )

            # Run schema migration
            subprocess.check_call(["dbos", "migrate"], cwd=temp_path, env=venv)

            # Launch the application in the virtual environment as a background process
            process = subprocess.Popen(["dbos", "start"], cwd=temp_path, env=venv)

            try:
                url = "http://localhost:8000"
                if db_starter:
                    url += "/greeting/dbos"
                max_retries = 10
                for attempt in range(max_retries):
                    try:
                        with urllib.request.urlopen(url, timeout=1) as response:
                            status_code = response.getcode()
                            assert status_code == 200
                            if db_starter:
                                response_data = response.read().decode("utf-8")
                                data = json.loads(response_data)
                                assert (
                                    data
                                    == "Greetings, dbos! You have been greeted 1 times."
                                )
                            break
                    except (urllib.error.URLError, AssertionError) as e:
                        if attempt < max_retries - 1:  # If not the last attempt
                            print(
                                f"Attempt {attempt + 1} failed: {e}. Retrying in 1 second..."
                            )
                            time.sleep(1)
                        else:
                            print(f"All {max_retries} attempts failed. Last error: {e}")
                            raise
            finally:
                os.kill(process.pid, signal.SIGINT)
                process.wait()


def test_init_config(skip_with_sqlite: None) -> None:
    app_name = "example-name"
    expected_yaml = {
        "name": app_name,
        "language": "python",
        "runtimeConfig": {"start": ["python3 main.py"]},
        "database_url": "${DBOS_DATABASE_URL}",
    }
    with tempfile.TemporaryDirectory() as temp_path:

        subprocess.check_call(
            ["dbos", "init", app_name, "--config"],
            cwd=temp_path,
        )

        config_path = os.path.join(temp_path, "dbos-config.yaml")
        assert os.path.exists(config_path)

        with open(config_path) as f:
            actual_yaml = yaml.safe_load(f)

        assert actual_yaml == expected_yaml


def test_reset(db_engine: sa.Engine, skip_with_sqlite: None) -> None:
    app_name = "reset-app"
    db_url = db_engine.url.set(database="reset_app").render_as_string(
        hide_password=False
    )
    sys_db_name = "reset_app_dbos_sys"
    sys_db_url = db_engine.url.set(database=sys_db_name).render_as_string(
        hide_password=False
    )
    with tempfile.TemporaryDirectory() as temp_path:
        env = os.environ.copy()
        env["DBOS_DATABASE_URL"] = db_url
        subprocess.check_call(
            ["dbos", "init", app_name, "--template", "dbos-db-starter"],
            cwd=temp_path,
        )

        # Create a system database and verify it exists
        subprocess.check_call(["dbos", "migrate"], cwd=temp_path, env=env)
        with db_engine.connect() as c:
            c.execution_options(isolation_level="AUTOCOMMIT")
            result = c.execute(
                sa.text(
                    f"SELECT COUNT(*) FROM pg_database WHERE datname = '{sys_db_name}'"
                )
            ).scalar()
            assert result == 1

        # Call reset and verify it's destroyed
        subprocess.check_call(
            ["dbos", "reset", "-y", "--db-url", db_url, "--sys-db-url", sys_db_url],
            cwd=temp_path,
        )
        with db_engine.connect() as c:
            c.execution_options(isolation_level="AUTOCOMMIT")
            result = c.execute(
                sa.text(
                    f"SELECT COUNT(*) FROM pg_database WHERE datname = '{sys_db_name}'"
                )
            ).scalar()
            assert result == 0


def test_workflow_commands(config: DBOSConfig) -> None:
    assert config["application_database_url"] is not None
    assert config["system_database_url"] is not None
    if using_sqlite():
        db_url = config["system_database_url"]
    else:
        db_url = (
            sa.make_url(config["system_database_url"])
            .set(database="dbos_toolbox")
            .render_as_string(hide_password=False)
        )
    with tempfile.TemporaryDirectory() as temp_path:
        env = os.environ.copy()
        env["DBOS_SYSTEM_DATABASE_URL"] = db_url
        subprocess.check_call(
            ["dbos", "init", "--template", "dbos-toolbox"],
            cwd=temp_path,
            env=env,
        )
        subprocess.check_call(
            ["dbos", "reset", "-y", "--sys-db-url", db_url], cwd=temp_path
        )

        schema = "dbos"
        # Get some workflows enqueued on the toolbox, then kill the toolbox
        process = subprocess.Popen(["python3", "main.py"], cwd=temp_path, env=env)
        try:
            session = requests.Session()
            for i in range(10):
                try:
                    session.get(
                        "http://localhost:8000/queue", timeout=1
                    ).raise_for_status()
                    break
                except requests.exceptions.Timeout:
                    break
                except requests.exceptions.ConnectionError as e:
                    if i == 9:
                        raise
                    print(f"Attempt {i+1} failed: {e}. Retrying in 1 second...")
                    time.sleep(1)
            time.sleep(1)  # So the queued workflows can start
        finally:
            # Because the toolbox steps sleep for 5 seconds, all the steps should be PENDING
            os.kill(process.pid, signal.SIGKILL)
            process.wait()

        # Verify the output is valid JSON
        output = subprocess.check_output(
            ["dbos", "workflow", "list", "--sys-db-url", db_url, "--schema", schema],
            cwd=temp_path,
        )
        data = json.loads(output)
        assert isinstance(data, list) and len(data) == 10

        # Verify the output is valid JSON
        output = subprocess.check_output(
            [
                "dbos",
                "workflow",
                "queue",
                "list",
                "--sys-db-url",
                db_url,
                "--schema",
                schema,
            ],
            cwd=temp_path,
        )
        workflows = json.loads(output)
        assert isinstance(workflows, list) and len(workflows) == 10
        for wf in workflows:
            output = subprocess.check_output(
                [
                    "dbos",
                    "workflow",
                    "get",
                    wf["workflow_id"],
                    "--sys-db-url",
                    db_url,
                    "--schema",
                    schema,
                ],
                cwd=temp_path,
            )
            get_wf_data = json.loads(output)
            assert isinstance(get_wf_data, dict)
            assert get_wf_data["workflow_id"] == wf["workflow_id"]

        # workflow ID is a preffix to each step ID
        wf_id = "-".join(workflows[0]["workflow_id"].split("-")[:-1])
        get_steps_output = subprocess.check_output(
            [
                "dbos",
                "workflow",
                "steps",
                wf_id,
                "--sys-db-url",
                db_url,
                "--schema",
                schema,
            ],
            cwd=temp_path,
        )
        get_steps_data = json.loads(get_steps_output)
        assert isinstance(get_steps_data, list)
        assert len(get_steps_data) == 10

        # From now pass database url in the environment

        # cancel the workflow and check the status is CANCELLED
        subprocess.check_output(
            ["dbos", "workflow", "cancel", wf_id, "--schema", schema],
            cwd=temp_path,
            env=env,
        )
        output = subprocess.check_output(
            ["dbos", "workflow", "get", wf_id, "--schema", schema],
            cwd=temp_path,
            env=env,
        )
        get_wf_data = json.loads(output)
        assert isinstance(get_wf_data, dict)
        assert get_wf_data["status"] == "CANCELLED"

        # resume the workflow and check the status is ENQUEUED
        subprocess.check_output(
            ["dbos", "workflow", "resume", wf_id, "--schema", schema],
            cwd=temp_path,
            env=env,
        )
        output = subprocess.check_output(
            ["dbos", "workflow", "get", wf_id, "--schema", schema],
            cwd=temp_path,
            env=env,
        )
        get_wf_data = json.loads(output)
        assert isinstance(get_wf_data, dict)
        assert get_wf_data["status"] == "ENQUEUED"

        # fork the workflow at step 5 and check it has a new ID and its status is ENQUEUED
        output = subprocess.check_output(
            ["dbos", "workflow", "fork", wf_id, "--step", "5", "--schema", schema],
            cwd=temp_path,
            env=env,
        )
        fork_wf_data = json.loads(output)
        assert isinstance(fork_wf_data, dict)
        assert fork_wf_data["workflow_id"] != wf_id
        assert fork_wf_data["status"] == "ENQUEUED"

        # fork the workflow with custom forked workflow ID
        custom_fork_id = "custom-fork-id-12345"
        output = subprocess.check_output(
            [
                "dbos",
                "workflow",
                "fork",
                wf_id,
                "--step",
                "3",
                "--forked-workflow-id",
                custom_fork_id,
            ],
            cwd=temp_path,
            env=env,
        )
        custom_fork_data = json.loads(output)
        assert isinstance(custom_fork_data, dict)
        assert custom_fork_data["workflow_id"] == custom_fork_id
        assert custom_fork_data["status"] == "ENQUEUED"

        # verify the forked workflow data with get command
        output = subprocess.check_output(
            ["dbos", "workflow", "get", custom_fork_id], cwd=temp_path, env=env
        )
        custom_fork_get_data = json.loads(output)
        assert isinstance(custom_fork_get_data, dict)
        assert custom_fork_get_data["workflow_id"] == custom_fork_id

        # fork the workflow with custom application version
        output = subprocess.check_output(
            [
                "dbos",
                "workflow",
                "fork",
                wf_id,
                "--step",
                "2",
                "--application-version",
                "test-version",
            ],
            cwd=temp_path,
            env=env,
        )
        version_fork_data = json.loads(output)
        assert isinstance(version_fork_data, dict)
        assert version_fork_data["workflow_id"] != wf_id
        assert version_fork_data["status"] == "ENQUEUED"

        # verify the forked workflow data with get command and check application version
        output = subprocess.check_output(
            ["dbos", "workflow", "get", version_fork_data["workflow_id"]],
            cwd=temp_path,
            env=env,
        )
        version_fork_get_data = json.loads(output)
        assert isinstance(version_fork_get_data, dict)
        assert version_fork_get_data["workflow_id"] == version_fork_data["workflow_id"]
        assert version_fork_get_data["app_version"] == "test-version"

        # fork the workflow with both custom ID and application version
        custom_fork_id2 = "custom-fork-with-version-67890"
        output = subprocess.check_output(
            [
                "dbos",
                "workflow",
                "fork",
                wf_id,
                "--step",
                "4",
                "--forked-workflow-id",
                custom_fork_id2,
                "--application-version",
                "v2.0.0",
            ],
            cwd=temp_path,
            env=env,
        )
        combined_fork_data = json.loads(output)
        assert isinstance(combined_fork_data, dict)
        assert combined_fork_data["workflow_id"] == custom_fork_id2
        assert combined_fork_data["status"] == "ENQUEUED"

        # verify the forked workflow data with get command and check both ID and application version
        output = subprocess.check_output(
            ["dbos", "workflow", "get", custom_fork_id2], cwd=temp_path, env=env
        )
        combined_fork_get_data = json.loads(output)
        assert isinstance(combined_fork_get_data, dict)
        assert combined_fork_get_data["workflow_id"] == custom_fork_id2
        assert combined_fork_get_data["app_version"] == "v2.0.0"



import socket
import threading
import time
import traceback
import uuid
from importlib.metadata import version
from typing import TYPE_CHECKING, Optional

from websockets import ConnectionClosed, ConnectionClosedOK, InvalidStatus
from websockets.sync.client import connect
from websockets.sync.connection import Connection

from dbos._context import SetWorkflowID
from dbos._utils import GlobalParams
from dbos._workflow_commands import (
    garbage_collect,
    get_workflow,
    global_timeout,
    list_queued_workflows,
    list_workflow_steps,
    list_workflows,
)

from . import protocol as p

if TYPE_CHECKING:
    from dbos import DBOS

ws_version = version("websockets")
use_keepalive = ws_version < "15.0"


class ConductorWebsocket(threading.Thread):

    def __init__(
        self, dbos: "DBOS", conductor_url: str, conductor_key: str, evt: threading.Event
    ):
        super().__init__(daemon=True)
        self.websocket: Optional[Connection] = None
        self.evt = evt
        self.dbos = dbos
        self.app_name = dbos._config["name"]
        self.url = (
            conductor_url.rstrip("/") + f"/websocket/{self.app_name}/{conductor_key}"
        )
        # TODO: once we can upgrade to websockets>=15.0, we can always use the built-in keepalive
        self.ping_interval = 20  # Time between pings in seconds
        self.ping_timeout = 15  # Time to wait for a pong response in seconds
        self.keepalive_thread: Optional[threading.Thread] = None
        self.pong_event: Optional[threading.Event] = None
        self.last_ping_time = -1.0

        self.dbos.logger.debug(
            f"Connecting to conductor at {self.url} using websockets version {ws_version}"
        )

    def keepalive(self) -> None:
        self.dbos.logger.debug("Starting keepalive thread")
        while not self.evt.is_set():
            if self.websocket is None or self.websocket.close_code is not None:
                time.sleep(1)
                continue
            try:
                self.last_ping_time = time.time()
                self.pong_event = self.websocket.ping()
                self.dbos.logger.debug("> Sent ping to conductor")
                pong_result = self.pong_event.wait(self.ping_timeout)
                elapsed_time = time.time() - self.last_ping_time
                if not pong_result:
                    self.dbos.logger.warning(
                        f"Failed to receive pong from conductor after {elapsed_time:.2f} seconds. Reconnecting."
                    )
                    self.websocket.close()
                    continue
                # Wait for the next ping interval
                self.dbos.logger.debug(
                    f"< Received pong from conductor after {elapsed_time:.2f} seconds"
                )
                wait_time = self.ping_interval - elapsed_time
                self.evt.wait(max(0, wait_time))
            except ConnectionClosed:
                # The main loop will try to reconnect
                self.dbos.logger.debug("Connection to conductor closed.")
            except Exception as e:
                self.dbos.logger.warning(f"Failed to send ping to conductor: {e}.")
                self.websocket.close()

    def run(self) -> None:
        while not self.evt.is_set():
            try:
                with connect(
                    self.url,
                    open_timeout=5,
                    close_timeout=5,
                    logger=self.dbos.logger,
                ) as websocket:
                    self.websocket = websocket
                    if use_keepalive and self.keepalive_thread is None:
                        self.keepalive_thread = threading.Thread(
                            target=self.keepalive,
                            daemon=True,
                        )
                        self.keepalive_thread.start()
                    while not self.evt.is_set():
                        message = websocket.recv()
                        if not isinstance(message, str):
                            self.dbos.logger.warning(
                                "Received unexpected non-str message"
                            )
                            continue
                        base_message = p.BaseMessage.from_json(message)
                        msg_type = base_message.type
                        error_message = None
                        if msg_type == p.MessageType.EXECUTOR_INFO:
                            info_response = p.ExecutorInfoResponse(
                                type=p.MessageType.EXECUTOR_INFO,
                                request_id=base_message.request_id,
                                executor_id=GlobalParams.executor_id,
                                application_version=GlobalParams.app_version,
                                hostname=socket.gethostname(),
                            )
                            websocket.send(info_response.to_json())
                            self.dbos.logger.info("Connected to DBOS conductor")
                        elif msg_type == p.MessageType.RECOVERY:
                            recovery_message = p.RecoveryRequest.from_json(message)
                            success = True
                            try:
                                self.dbos._recover_pending_workflows(
                                    recovery_message.executor_ids
                                )
                            except Exception as e:
                                error_message = f"Exception encountered when recovering workflows: {traceback.format_exc()}"
                                self.dbos.logger.error(error_message)
                                success = False
                            recovery_response = p.RecoveryResponse(
                                type=p.MessageType.RECOVERY,
                                request_id=base_message.request_id,
                                success=success,
                                error_message=error_message,
                            )
                            websocket.send(recovery_response.to_json())
                        elif msg_type == p.MessageType.CANCEL:
                            cancel_message = p.CancelRequest.from_json(message)
                            success = True
                            try:
                                self.dbos.cancel_workflow(cancel_message.workflow_id)
                            except Exception as e:
                                error_message = f"Exception encountered when cancelling workflow {cancel_message.workflow_id}: {traceback.format_exc()}"
                                self.dbos.logger.error(error_message)
                                success = False
                            cancel_response = p.CancelResponse(
                                type=p.MessageType.CANCEL,
                                request_id=base_message.request_id,
                                success=success,
                                error_message=error_message,
                            )
                            websocket.send(cancel_response.to_json())
                        elif msg_type == p.MessageType.RESUME:
                            resume_message = p.ResumeRequest.from_json(message)
                            success = True
                            try:
                                self.dbos.resume_workflow(resume_message.workflow_id)
                            except Exception as e:
                                error_message = f"Exception encountered when resuming workflow {resume_message.workflow_id}: {traceback.format_exc()}"
                                self.dbos.logger.error(error_message)
                                success = False
                            resume_response = p.ResumeResponse(
                                type=p.MessageType.RESUME,
                                request_id=base_message.request_id,
                                success=success,
                                error_message=error_message,
                            )
                            websocket.send(resume_response.to_json())
                        elif msg_type == p.MessageType.RESTART:
                            # TODO: deprecate this message type in favor of Fork
                            restart_message = p.RestartRequest.from_json(message)
                            success = True
                            try:
                                self.dbos.fork_workflow(restart_message.workflow_id, 1)
                            except Exception as e:
                                error_message = f"Exception encountered when restarting workflow {restart_message.workflow_id}: {traceback.format_exc()}"
                                self.dbos.logger.error(error_message)
                                success = False
                            restart_response = p.RestartResponse(
                                type=p.MessageType.RESTART,
                                request_id=base_message.request_id,
                                success=success,
                                error_message=error_message,
                            )
                            websocket.send(restart_response.to_json())
                        elif msg_type == p.MessageType.FORK_WORKFLOW:
                            fork_message = p.ForkWorkflowRequest.from_json(message)
                            new_workflow_id = fork_message.body["new_workflow_id"]
                            if new_workflow_id is None:
                                new_workflow_id = str(uuid.uuid4())
                            workflow_id = fork_message.body["workflow_id"]
                            start_step = fork_message.body["start_step"]
                            app_version = fork_message.body["application_version"]
                            try:
                                with SetWorkflowID(new_workflow_id):
                                    new_handle = self.dbos.fork_workflow(
                                        workflow_id,
                                        start_step,
                                        application_version=app_version,
                                    )
                                new_workflow_id = new_handle.workflow_id
                            except Exception as e:
                                error_message = f"Exception encountered when forking workflow {workflow_id} to new workflow {new_workflow_id} on step {start_step}, app version {app_version}: {traceback.format_exc()}"
                                self.dbos.logger.error(error_message)
                                new_workflow_id = None

                            fork_response = p.ForkWorkflowResponse(
                                type=p.MessageType.FORK_WORKFLOW,
                                request_id=base_message.request_id,
                                new_workflow_id=new_workflow_id,
                                error_message=error_message,
                            )
                            websocket.send(fork_response.to_json())
                        elif msg_type == p.MessageType.LIST_WORKFLOWS:
                            list_workflows_message = p.ListWorkflowsRequest.from_json(
                                message
                            )
                            body = list_workflows_message.body
                            infos = []
                            try:
                                load_input = body.get("load_input", False)
                                load_output = body.get("load_output", False)
                                infos = list_workflows(
                                    self.dbos._sys_db,
                                    workflow_ids=body["workflow_uuids"],
                                    user=body["authenticated_user"],
                                    start_time=body["start_time"],
                                    end_time=body["end_time"],
                                    status=body["status"],
                                    app_version=body["application_version"],
                                    name=body["workflow_name"],
                                    limit=body["limit"],
                                    offset=body["offset"],
                                    sort_desc=body["sort_desc"],
                                    load_input=load_input,
                                    load_output=load_output,
                                )
                            except Exception as e:
                                error_message = f"Exception encountered when listing workflows: {traceback.format_exc()}"
                                self.dbos.logger.error(error_message)

                            list_workflows_response = p.ListWorkflowsResponse(
                                type=p.MessageType.LIST_WORKFLOWS,
                                request_id=base_message.request_id,
                                output=[
                                    p.WorkflowsOutput.from_workflow_information(i)
                                    for i in infos
                                ],
                                error_message=error_message,
                            )
                            websocket.send(list_workflows_response.to_json())
                        elif msg_type == p.MessageType.LIST_QUEUED_WORKFLOWS:
                            list_queued_workflows_message = (
                                p.ListQueuedWorkflowsRequest.from_json(message)
                            )
                            q_body = list_queued_workflows_message.body
                            infos = []
                            try:
                                q_load_input = q_body.get("load_input", False)
                                infos = list_queued_workflows(
                                    self.dbos._sys_db,
                                    start_time=q_body["start_time"],
                                    end_time=q_body["end_time"],
                                    status=q_body["status"],
                                    name=q_body["workflow_name"],
                                    limit=q_body["limit"],
                                    offset=q_body["offset"],
                                    queue_name=q_body["queue_name"],
                                    sort_desc=q_body["sort_desc"],
                                    load_input=q_load_input,
                                )
                            except Exception as e:
                                error_message = f"Exception encountered when listing queued workflows: {traceback.format_exc()}"
                                self.dbos.logger.error(error_message)

                            list_queued_workflows_response = (
                                p.ListQueuedWorkflowsResponse(
                                    type=p.MessageType.LIST_QUEUED_WORKFLOWS,
                                    request_id=base_message.request_id,
                                    output=[
                                        p.WorkflowsOutput.from_workflow_information(i)
                                        for i in infos
                                    ],
                                    error_message=error_message,
                                )
                            )
                            websocket.send(list_queued_workflows_response.to_json())
                        elif msg_type == p.MessageType.GET_WORKFLOW:
                            get_workflow_message = p.GetWorkflowRequest.from_json(
                                message
                            )
                            info = None
                            try:
                                info = get_workflow(
                                    self.dbos._sys_db, get_workflow_message.workflow_id
                                )
                            except Exception as e:
                                error_message = f"Exception encountered when getting workflow {get_workflow_message.workflow_id}: {traceback.format_exc()}"
                                self.dbos.logger.error(error_message)

                            get_workflow_response = p.GetWorkflowResponse(
                                type=p.MessageType.GET_WORKFLOW,
                                request_id=base_message.request_id,
                                output=(
                                    p.WorkflowsOutput.from_workflow_information(info)
                                    if info is not None
                                    else None
                                ),
                                error_message=error_message,
                            )
                            websocket.send(get_workflow_response.to_json())
                        elif msg_type == p.MessageType.EXIST_PENDING_WORKFLOWS:
                            exist_pending_workflows_message = (
                                p.ExistPendingWorkflowsRequest.from_json(message)
                            )
                            pending_wfs = []
                            try:
                                pending_wfs = self.dbos._sys_db.get_pending_workflows(
                                    exist_pending_workflows_message.executor_id,
                                    exist_pending_workflows_message.application_version,
                                )
                            except Exception as e:
                                error_message = f"Exception encountered when checking for pending workflows: {traceback.format_exc()}"
                                self.dbos.logger.error(error_message)

                            exist_pending_workflows_response = (
                                p.ExistPendingWorkflowsResponse(
                                    type=p.MessageType.EXIST_PENDING_WORKFLOWS,
                                    request_id=base_message.request_id,
                                    exist=len(pending_wfs) > 0,
                                    error_message=error_message,
                                )
                            )
                            websocket.send(exist_pending_workflows_response.to_json())
                        elif msg_type == p.MessageType.LIST_STEPS:
                            list_steps_message = p.ListStepsRequest.from_json(message)
                            step_info = None
                            try:
                                step_info = list_workflow_steps(
                                    self.dbos._sys_db,
                                    self.dbos._app_db,
                                    list_steps_message.workflow_id,
                                )
                            except Exception as e:
                                error_message = f"Exception encountered when getting workflow {list_steps_message.workflow_id}: {traceback.format_exc()}"
                                self.dbos.logger.error(error_message)

                            list_steps_response = p.ListStepsResponse(
                                type=p.MessageType.LIST_STEPS,
                                request_id=base_message.request_id,
                                output=(
                                    [
                                        p.WorkflowSteps.from_step_info(i)
                                        for i in step_info
                                    ]
                                    if step_info is not None
                                    else None
                                ),
                                error_message=error_message,
                            )
                            websocket.send(list_steps_response.to_json())
                        elif msg_type == p.MessageType.RETENTION:
                            retention_message = p.RetentionRequest.from_json(message)
                            success = True
                            try:
                                garbage_collect(
                                    self.dbos,
                                    cutoff_epoch_timestamp_ms=retention_message.body[
                                        "gc_cutoff_epoch_ms"
                                    ],
                                    rows_threshold=retention_message.body[
                                        "gc_rows_threshold"
                                    ],
                                )
                                if (
                                    retention_message.body["timeout_cutoff_epoch_ms"]
                                    is not None
                                ):
                                    global_timeout(
                                        self.dbos,
                                        retention_message.body[
                                            "timeout_cutoff_epoch_ms"
                                        ],
                                    )
                            except Exception as e:
                                error_message = f"Exception encountered during enforcing retention policy: {traceback.format_exc()}"
                                self.dbos.logger.error(error_message)
                                success = False

                            retention_response = p.RetentionResponse(
                                type=p.MessageType.RETENTION,
                                request_id=base_message.request_id,
                                success=success,
                                error_message=error_message,
                            )
                            websocket.send(retention_response.to_json())
                        else:
                            self.dbos.logger.warning(
                                f"Unexpected message type: {msg_type}"
                            )
                            unknown_message = p.BaseResponse(
                                request_id=base_message.request_id,
                                type=msg_type,
                                error_message="Unknown message type",
                            )
                            # Still need to send a response to the conductor
                            websocket.send(unknown_message.to_json())
            except ConnectionClosedOK:
                if self.evt.is_set():
                    self.dbos.logger.info("Conductor connection terminated")
                    break
                # Otherwise, we are trying to reconnect
                self.dbos.logger.warning(
                    "Connection to conductor lost. Reconnecting..."
                )
                time.sleep(1)
                continue
            except ConnectionClosed as e:
                self.dbos.logger.warning(
                    f"Connection to conductor lost. Reconnecting: {e}"
                )
                time.sleep(1)
                continue
            except InvalidStatus as e:
                # This happens when it cannot open a connection to the conductor. E.g., the conductor rejects the request
                json_data = e.response.body.decode("utf-8")
                self.dbos.logger.error(
                    f"Failed to connect to conductor. Retrying: {str(e) }. Details: {json_data}"
                )
                time.sleep(1)
                continue
            except Exception as e:
                self.dbos.logger.error(
                    f"Unexpected exception in connection to conductor. Reconnecting: {e}"
                )
                time.sleep(1)
                continue

        # Wait for the keepalive thread to finish
        if self.keepalive_thread is not None:
            if self.pong_event is not None:
                self.pong_event.set()
            self.keepalive_thread.join()



import os
import socket
import threading
import time
import uuid
from datetime import datetime, timedelta, timezone
from typing import Any, Dict

import pytest
import requests
import sqlalchemy as sa
from requests.exceptions import ConnectionError

# Public API
from dbos import (
    DBOS,
    DBOSConfig,
    Queue,
    SetWorkflowID,
    WorkflowHandle,
    _workflow_commands,
)
from dbos._error import DBOSAwaitedWorkflowCancelledError
from dbos._schemas.system_database import SystemSchema
from dbos._sys_db import WorkflowStatusString
from dbos._utils import INTERNAL_QUEUE_NAME, GlobalParams


def test_admin_endpoints(dbos: DBOS) -> None:

    # Test GET /dbos-healthz
    response = requests.get("http://localhost:3001/dbos-healthz", timeout=5)
    assert response.status_code == 200
    assert response.text == "healthy"

    # Test POST /dbos-workflow-recovery
    data = ["executor1", "executor2"]
    response = requests.post(
        "http://localhost:3001/dbos-workflow-recovery", json=data, timeout=5
    )
    assert response.status_code == 200
    assert response.json() == []

    # Test GET /dbos-workflow-queues-metadata
    Queue("q1")
    Queue("q2", concurrency=1)
    Queue("q3", concurrency=1, worker_concurrency=1)
    Queue("q4", concurrency=1, worker_concurrency=1, limiter={"limit": 0, "period": 0})
    response = requests.get(
        "http://localhost:3001/dbos-workflow-queues-metadata", timeout=5
    )
    assert response.status_code == 200
    assert response.json() == [
        {"name": INTERNAL_QUEUE_NAME},
        {"name": "q1"},
        {"name": "q2", "concurrency": 1},
        {"name": "q3", "concurrency": 1, "workerConcurrency": 1},
        {
            "name": "q4",
            "concurrency": 1,
            "workerConcurrency": 1,
            "rateLimit": {"limit": 0, "period": 0},
        },
    ]

    # Test GET not found
    response = requests.get("http://localhost:3001/stuff", timeout=5)
    assert response.status_code == 404

    # Test POST not found
    response = requests.post("http://localhost:3001/stuff", timeout=5)
    assert response.status_code == 404

    response = requests.get("http://localhost:3001/deactivate", timeout=5)
    assert response.status_code == 200

    for event in dbos.poller_stop_events:
        assert event.is_set()


def test_deactivate(dbos: DBOS, config: DBOSConfig) -> None:
    wf_counter: int = 0

    queue = Queue("example-queue")

    @DBOS.scheduled("* * * * * *")
    @DBOS.workflow()
    def test_workflow(scheduled: datetime, actual: datetime) -> None:
        nonlocal wf_counter
        wf_counter += 1

    @DBOS.workflow()
    def regular_workflow() -> int:
        return 5

    # Let the scheduled workflow run
    time.sleep(5)
    val = wf_counter
    assert val > 0
    # Deactivate--scheduled workflow should stop
    response = requests.get("http://localhost:3001/deactivate", timeout=5)
    assert response.status_code == 200
    for event in dbos.poller_stop_events:
        assert event.is_set()
    # Verify the scheduled workflow does not run anymore
    time.sleep(5)
    assert wf_counter <= val + 2
    # Enqueue a workflow, verify it still runs
    assert queue.enqueue(regular_workflow).get_result() == 5

    # Test deferred event receivers
    DBOS.destroy(destroy_registry=True)
    dbos = DBOS(config=config)

    @DBOS.scheduled("* * * * * *")
    @DBOS.workflow()
    def deferred_workflow(scheduled: datetime, actual: datetime) -> None:
        nonlocal wf_counter
        wf_counter += 1

    DBOS.launch()
    assert len(dbos.poller_stop_events) > 0
    for event in dbos.poller_stop_events:
        assert not event.is_set()
    # Deactivate--scheduled workflow should stop
    response = requests.get("http://localhost:3001/deactivate", timeout=5)
    assert response.status_code == 200
    for event in dbos.poller_stop_events:
        assert event.is_set()


def test_admin_recovery(config: DBOSConfig) -> None:
    os.environ["DBOS__VMID"] = "testexecutor"
    os.environ["DBOS__APPVERSION"] = "testversion"
    os.environ["DBOS__APPID"] = "testappid"
    DBOS.destroy(destroy_registry=True)
    dbos = DBOS(config=config)
    DBOS.launch()

    step_counter: int = 0
    wf_counter: int = 0

    @DBOS.workflow()
    def test_workflow(var: str, var2: str) -> str:
        nonlocal wf_counter
        wf_counter += 1
        res = test_step(var2)
        return res + var

    @DBOS.step()
    def test_step(var2: str) -> str:
        nonlocal step_counter
        step_counter += 1
        return var2 + "1"

    wfuuid = str(uuid.uuid4())
    DBOS.logger.info("Initiating: " + wfuuid)
    with SetWorkflowID(wfuuid):
        assert test_workflow("bob", "bob") == "bob1bob"

    # Manually update the database to pretend the workflow comes from another executor and is pending
    with dbos._sys_db.engine.begin() as c:
        query = (
            sa.update(SystemSchema.workflow_status)
            .values(
                status=WorkflowStatusString.PENDING.value, executor_id="other-executor"
            )
            .where(SystemSchema.workflow_status.c.workflow_uuid == wfuuid)
        )
        c.execute(query)

    status = dbos.get_workflow_status(wfuuid)
    assert (
        status is not None and status.status == "PENDING"
    ), "Workflow status not updated"

    # Test POST /dbos-workflow-recovery
    data = ["other-executor"]
    response = requests.post(
        "http://localhost:3001/dbos-workflow-recovery", json=data, timeout=5
    )
    assert response.status_code == 200
    assert response.json() == [wfuuid]

    # Wait until the workflow is recovered
    max_retries = 10
    succeeded = False
    for attempt in range(max_retries):
        status = dbos.get_workflow_status(wfuuid)
        if status is not None and status.status == "SUCCESS":
            assert status.status == "SUCCESS"
            assert status.executor_id == "testexecutor"
            succeeded = True
            break
        else:
            time.sleep(1)
            print(f"Attempt {attempt + 1} failed. Retrying in 1 second...")
    assert succeeded, "Workflow did not recover"


def test_admin_diff_port(config: DBOSConfig, cleanup_test_databases: None) -> None:
    # Initialize singleton
    DBOS.destroy()  # In case of other tests leaving it

    try:
        # Initialize DBOS
        config["admin_port"] = 8001
        DBOS(config=config)
        DBOS.launch()

        # Test GET /dbos-healthz
        response = requests.get("http://localhost:8001/dbos-healthz", timeout=5)
        assert response.status_code == 200
        assert response.text == "healthy"
    finally:
        # Clean up after the test
        DBOS.destroy()


def test_disable_admin_server(config: DBOSConfig, cleanup_test_databases: None) -> None:
    # Initialize singleton
    DBOS.destroy()  # In case of other tests leaving it

    config["run_admin_server"] = False
    try:
        DBOS(config=config)
        DBOS.launch()

        with pytest.raises(ConnectionError):
            requests.get("http://localhost:3001/dbos-healthz", timeout=1)
    finally:
        # Clean up after the test
        DBOS.destroy()


def test_busy_admin_server_port_does_not_throw(config: DBOSConfig) -> None:
    # Initialize singleton
    DBOS.destroy()  # In case of other tests leaving it

    server_thread = None
    stop_event = threading.Event()
    try:

        def start_dummy_server(port: int, stop_event: threading.Event) -> None:
            """Starts a simple TCP server on the given port."""
            server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            server_socket.bind(("0.0.0.0", port))
            server_socket.listen(1)
            # We need to call accept for the port to be considered busy by the OS...
            while not stop_event.is_set():
                try:
                    server_socket.settimeout(
                        1
                    )  # Timeout for accept in case the thread is stopped.
                    client_socket, _ = server_socket.accept()
                    client_socket.close()
                except socket.timeout:
                    pass
            server_socket.close()

        port_to_block = 3001
        server_thread = threading.Thread(
            target=start_dummy_server, args=(port_to_block, stop_event)
        )
        server_thread.daemon = (
            True  # Allows the thread to be terminated when the main thread exits
        )
        server_thread.start()

        DBOS(config=config)
        DBOS.launch()
    finally:
        # Clean up after the test
        DBOS.destroy()
        if server_thread and server_thread.is_alive():
            stop_event.set()
            server_thread.join(2)
            if server_thread.is_alive():
                print("Warning: Server thread did not terminate gracefully.")


def test_admin_workflow_resume(dbos: DBOS) -> None:
    event = threading.Event()

    @DBOS.workflow()
    def blocking_workflow() -> None:
        event.wait()
        DBOS.sleep(0.1)

    # Run the workflow
    handle = DBOS.start_workflow(blocking_workflow)

    # Verify the workflow is pending
    output = DBOS.list_workflows()
    assert len(output) == 1
    assert output[0] != None
    wfid = output[0].workflow_id
    info = _workflow_commands.get_workflow(dbos._sys_db, wfid)
    assert info is not None
    assert info.status == "PENDING"

    # Cancel the workflow. Verify it was cancelled
    response = requests.post(
        f"http://localhost:3001/workflows/{wfid}/cancel", json=[], timeout=5
    )
    assert response.status_code == 204
    event.set()
    with pytest.raises(DBOSAwaitedWorkflowCancelledError):
        handle.get_result()
    info = _workflow_commands.get_workflow(dbos._sys_db, wfid)
    assert info is not None
    assert info.status == "CANCELLED"

    # Manually update the database to pretend the workflow comes from another executor and is pending
    with dbos._sys_db.engine.begin() as c:
        query = (
            sa.update(SystemSchema.workflow_status)
            .values(
                status=WorkflowStatusString.PENDING.value, executor_id="other-executor"
            )
            .where(SystemSchema.workflow_status.c.workflow_uuid == wfid)
        )
        c.execute(query)

    # Resume the workflow. Verify that it succeeds again.
    response = requests.post(
        f"http://localhost:3001/workflows/{wfid}/resume", json=[], timeout=5
    )
    assert response.status_code == 204
    # Wait for the workflow to finish
    DBOS.retrieve_workflow(wfid).get_result()
    info = _workflow_commands.get_workflow(dbos._sys_db, wfid)
    assert info is not None
    assert info.status == "SUCCESS"
    assert info.executor_id == GlobalParams.executor_id


def test_admin_workflow_restart(dbos: DBOS) -> None:

    @DBOS.workflow()
    def simple_workflow() -> None:
        print("Executed Simple workflow")
        return

    # run the workflow
    simple_workflow()
    time.sleep(1)

    # get the workflow list
    output = DBOS.list_workflows()
    assert len(output) == 1, f"Expected list length to be 1, but got {len(output)}"

    assert output[0] != None, "Expected output to be not None"

    wfUuid = output[0].workflow_id

    info = _workflow_commands.get_workflow(dbos._sys_db, wfUuid)
    assert info is not None, "Expected output to be not None"

    assert info.status == "SUCCESS", f"Expected status to be SUCCESS"

    response = requests.post(
        f"http://localhost:3001/workflows/{wfUuid}/restart", json=[], timeout=5
    )
    assert response.status_code == 200

    new_workflow_id = response.json().get("workflow_id")
    print(f"New workflow ID: {new_workflow_id}")
    assert new_workflow_id is not None, "Expected new workflow ID to be not None"

    worked = False
    count = 0
    while count < 5:
        # Check if the workflow is in the database
        info = _workflow_commands.get_workflow(dbos._sys_db, new_workflow_id)
        assert info is not None, "Expected output to be not None"
        if info.status == "SUCCESS":
            worked = True
            break
        time.sleep(1)
        count += 1

    assert worked, "Workflow did not finish successfully"


def test_admin_workflow_fork(dbos: DBOS) -> None:

    @DBOS.workflow()
    def simple_workflow() -> None:
        print("Executed Simple workflow")
        return

    # run the workflow
    simple_workflow()

    # get the workflow list
    output = DBOS.list_workflows()
    assert len(output) == 1, f"Expected list length to be 1, but got {len(output)}"

    assert output[0] != None, "Expected output to be not None"

    wfUuid = output[0].workflow_id

    info = _workflow_commands.get_workflow(dbos._sys_db, wfUuid)
    assert info is not None, "Expected output to be not None"

    assert info.status == "SUCCESS", f"Expected status to be SUCCESS"

    response = requests.post(
        f"http://localhost:3001/workflows/{wfUuid}/fork", json={}, timeout=5
    )
    assert response.status_code == 200

    new_workflow_id = response.json().get("workflow_id")
    print(f"New workflow ID: {new_workflow_id}")
    assert new_workflow_id is not None, "Expected new workflow ID to be not None"
    assert new_workflow_id != wfUuid, "Expected new workflow ID to be different"

    output = DBOS.list_workflows()
    assert len(output) == 2, f"Expected list length to be 2, but got {len(output)}"

    worked = False
    count = 0
    while count < 5:
        # Check if the workflow is in the database
        info = _workflow_commands.get_workflow(dbos._sys_db, new_workflow_id)
        assert info is not None, "Expected output to be not None"
        if info.status == "SUCCESS":
            worked = True
            break
        time.sleep(1)
        count += 1

    # test for new_workflow_id and app version

    new_version = "my_new_version"
    GlobalParams.app_version = new_version

    response = requests.post(
        f"http://localhost:3001/workflows/{wfUuid}/fork",
        json={"new_workflow_id": "123456", "application_version": new_version},
        timeout=5,
    )
    assert response.status_code == 200

    new_workflow_id = response.json().get("workflow_id")
    assert new_workflow_id == "123456", "Expected new workflow ID is not 123456"

    handle: WorkflowHandle[None] = dbos.retrieve_workflow(new_workflow_id)
    assert (
        handle.get_status().app_version == new_version
    ), f"Expected application version to be {new_version}, but got {handle.get_status().app_version}"

    assert worked, "Workflow did not finish successfully"


def test_list_workflows(dbos: DBOS, skip_with_sqlite_imprecise_time: None) -> None:
    # Create workflows for testing
    @DBOS.workflow()
    def test_workflow_1() -> None:
        pass

    @DBOS.workflow()
    def test_workflow_2(my_time: datetime) -> str:
        workflow_id = DBOS.workflow_id
        assert workflow_id is not None
        return workflow_id + " completed at " + my_time.isoformat()

    # Start workflows
    handle_1 = DBOS.start_workflow(test_workflow_1)
    time.sleep(2)  # Sleep for 2 seconds between workflows
    handle_2 = DBOS.start_workflow(test_workflow_2, datetime.now())

    # Wait for workflows to complete
    handle_1.get_result()
    handle_2.get_result()

    # List workflows and dynamically set the filter "name"
    workflows_list = DBOS.list_workflows()
    assert (
        len(workflows_list) >= 2
    ), f"Expected at least 2 workflows, but got {len(workflows_list)}"

    workflow_ids = [w.workflow_id for w in workflows_list]

    # Convert created_at to ISO 8601 format
    created_at_second_workflow = workflows_list[1].created_at
    assert (
        created_at_second_workflow is not None
    ), "created_at for the second workflow is None"
    start_time_filter = datetime.fromtimestamp(
        created_at_second_workflow / 1000, tz=timezone.utc
    ).isoformat()

    # Test POST /workflows with filters
    filters: Dict[str, Any] = {
        "workflow_uuids": workflow_ids,
        "start_time": start_time_filter,
    }
    response = requests.post("http://localhost:3001/workflows", json=filters, timeout=5)
    assert response.status_code == 200

    workflows = response.json()
    assert len(workflows) == 1, f"Expected 1 workflows, but got {len(workflows)}"

    # Make sure it contains all the expected fields
    assert workflows[0]["WorkflowUUID"] == handle_2.workflow_id, "Workflow ID mismatch"
    assert workflows[0]["WorkflowName"] == test_workflow_2.__qualname__
    assert workflows[0]["Status"] == "SUCCESS"
    assert workflows[0]["WorkflowClassName"] is None
    assert workflows[0]["WorkflowConfigName"] is None
    assert workflows[0]["AuthenticatedUser"] is None
    assert workflows[0]["AssumedRole"] is None
    assert workflows[0]["AuthenticatedRoles"] is None
    # By default, input and output are not loaded
    assert workflows[0]["Input"] is None
    assert workflows[0]["Output"] is None
    assert workflows[0]["Error"] is None
    assert workflows[0]["CreatedAt"] is not None and len(workflows[0]["CreatedAt"]) > 0
    assert workflows[0]["UpdatedAt"] is not None and len(workflows[0]["UpdatedAt"]) > 0
    assert workflows[0]["QueueName"] is None
    assert workflows[0]["ApplicationVersion"] == DBOS.application_version
    assert workflows[0]["ExecutorID"] == GlobalParams.executor_id

    # Only load input and output as requested
    filters = {
        "workflow_uuids": workflow_ids,
        "start_time": start_time_filter,
        "load_input": True,
        "load_output": True,
    }
    response = requests.post("http://localhost:3001/workflows", json=filters, timeout=5)
    assert response.status_code == 200

    workflows = response.json()
    assert len(workflows) == 1
    assert workflows[0]["Output"] is not None and len(workflows[0]["Output"]) > 0
    assert workflows[0]["Input"] is not None and len(workflows[0]["Input"]) > 0

    # Test POST /workflows without filters
    response = requests.post("http://localhost:3001/workflows", json={}, timeout=5)
    assert response.status_code == 200

    workflows = response.json()
    assert len(workflows) == len(
        workflows_list
    ), f"Expected {len(workflows_list)} workflows, but got {len(workflows)}"
    for workflow in workflows:
        assert workflow["WorkflowUUID"] in workflow_ids, "Workflow ID mismatch"

    # Verify sort_desc inverts the order
    filters = {
        "sort_desc": True,
    }
    response = requests.post("http://localhost:3001/workflows", json=filters, timeout=5)
    assert response.status_code == 200
    workflows = response.json()
    assert len(workflows) == len(workflows_list)
    assert (
        workflows[0]["WorkflowUUID"] == handle_2.workflow_id
    ), "First workflow should be the last one started"

    # Test all filters
    filters = {
        "workflow_uuids": ["not-a-valid-uuid"],
    }
    response = requests.post("http://localhost:3001/workflows", json=filters, timeout=5)
    assert response.status_code == 200
    workflows = response.json()
    assert len(workflows) == 0, "Expected no workflows for invalid UUID"

    filters = {
        "workflow_uuids": [handle_1.workflow_id, handle_2.workflow_id],
    }
    response = requests.post("http://localhost:3001/workflows", json=filters, timeout=5)
    assert response.status_code == 200
    workflows = response.json()
    assert len(workflows) == 2

    filters = {
        "authenticated_user": "no-user",
    }
    response = requests.post("http://localhost:3001/workflows", json=filters, timeout=5)
    assert response.status_code == 200
    workflows = response.json()
    assert len(workflows) == 0

    filters = {
        "workflow_name": test_workflow_1.__qualname__,
    }
    response = requests.post("http://localhost:3001/workflows", json=filters, timeout=5)
    assert response.status_code == 200
    workflows = response.json()
    assert len(workflows) == 1
    assert workflows[0]["WorkflowUUID"] == handle_1.workflow_id

    filters = {
        "end_time": (datetime.now(timezone.utc) - timedelta(minutes=10)).isoformat()
    }
    response = requests.post("http://localhost:3001/workflows", json=filters, timeout=5)
    assert response.status_code == 200
    workflows = response.json()
    assert len(workflows) == 0

    filters = {
        "start_time": (datetime.now(timezone.utc) + timedelta(minutes=10)).isoformat(),
    }
    response = requests.post("http://localhost:3001/workflows", json=filters, timeout=5)
    assert response.status_code == 200
    workflows = response.json()
    assert len(workflows) == 0

    filters = {
        "status": ["SUCCESS", "CANCELLED"],
    }
    response = requests.post("http://localhost:3001/workflows", json=filters, timeout=5)
    assert response.status_code == 200
    workflows = response.json()
    assert len(workflows) == 2

    filters = {
        "application_version": DBOS.application_version,
    }
    response = requests.post("http://localhost:3001/workflows", json=filters, timeout=5)
    assert response.status_code == 200
    workflows = response.json()
    assert len(workflows) == 2

    filters = {
        "limit": 1,
        "offset": 1,
    }
    response = requests.post("http://localhost:3001/workflows", json=filters, timeout=5)
    assert response.status_code == 200
    workflows = response.json()
    assert len(workflows) == 1
    assert workflows[0]["WorkflowUUID"] == handle_2.workflow_id

    filters = {
        "workflow_id_prefix": handle_1.workflow_id[
            :10
        ],  # First 10 characters of the workflow name
    }
    response = requests.post("http://localhost:3001/workflows", json=filters, timeout=5)
    assert response.status_code == 200
    workflows = response.json()
    assert len(workflows) == 1
    assert workflows[0]["WorkflowUUID"] == handle_1.workflow_id


def test_get_workflow_by_id(dbos: DBOS) -> None:
    # Create workflows for testing
    @DBOS.workflow()
    def test_workflow_1() -> None:
        pass

    @DBOS.workflow()
    def test_workflow_2(my_time: datetime) -> str:
        workflow_id = DBOS.workflow_id
        assert workflow_id is not None
        return workflow_id + " completed at " + my_time.isoformat()

    # Start workflows
    handle_1 = DBOS.start_workflow(test_workflow_1)
    handle_2 = DBOS.start_workflow(test_workflow_2, datetime.now())

    # Wait for workflows to complete
    handle_1.get_result()
    assert handle_2.get_result() is not None

    # Get the workflow ID of the second workflow
    workflow_id = handle_2.workflow_id

    # Test GET /workflows/:workflow_id for an existing workflow
    response = requests.get(f"http://localhost:3001/workflows/{workflow_id}", timeout=5)
    assert (
        response.status_code == 200
    ), f"Expected status code 200, but got {response.status_code}"

    workflow_data = response.json()
    assert workflow_data["WorkflowUUID"] == workflow_id, "Workflow ID mismatch"
    assert (
        workflow_data["Status"] == "SUCCESS"
    ), "Expected workflow status to be SUCCESS"
    assert workflow_data["WorkflowName"] == test_workflow_2.__qualname__
    assert workflow_data["WorkflowClassName"] is None
    assert workflow_data["WorkflowConfigName"] is None
    assert workflow_data["AuthenticatedUser"] is None
    assert workflow_data["AssumedRole"] is None
    assert workflow_data["AuthenticatedRoles"] is None
    assert workflow_data["Input"] is not None and len(workflow_data["Input"]) > 0
    assert workflow_data["Output"] is not None and len(workflow_data["Output"]) > 0
    assert workflow_data["Error"] is None
    assert (
        workflow_data["CreatedAt"] is not None and len(workflow_data["CreatedAt"]) > 0
    )
    assert (
        workflow_data["UpdatedAt"] is not None and len(workflow_data["UpdatedAt"]) > 0
    )
    assert workflow_data["QueueName"] is None
    assert workflow_data["ApplicationVersion"] == DBOS.application_version
    assert workflow_data["ExecutorID"] == GlobalParams.executor_id

    # Test GET /workflows/:workflow_id for a non-existing workflow
    non_existing_workflow_id = "non-existing-id"
    response = requests.get(
        f"http://localhost:3001/workflows/{non_existing_workflow_id}", timeout=5
    )
    assert (
        response.status_code == 404
    ), f"Expected status code 404, but got {response.status_code}"


def test_admin_garbage_collect(dbos: DBOS) -> None:

    @DBOS.workflow()
    def workflow() -> str:
        workflow_id = DBOS.workflow_id
        assert workflow_id is not None
        return workflow_id

    workflow()

    assert len(DBOS.list_workflows()) == 1

    response = requests.post(
        f"http://localhost:3001/dbos-garbage-collect",
        json={"cutoff_epoch_timestamp_ms": int(time.time() * 1000)},
        timeout=5,
    )
    response.raise_for_status()

    assert len(DBOS.list_workflows()) == 0


def test_admin_global_timeout(dbos: DBOS) -> None:

    @DBOS.workflow()
    def workflow() -> None:
        while True:
            DBOS.sleep(0.1)

    handle = DBOS.start_workflow(workflow)
    time.sleep(1)
    cutoff_epoch_timestamp_ms = int(time.time() * 1000) - 1000
    response = requests.post(
        f"http://localhost:3001/dbos-global-timeout",
        json={"cutoff_epoch_timestamp_ms": cutoff_epoch_timestamp_ms},
        timeout=5,
    )
    response.raise_for_status()
    with pytest.raises(DBOSAwaitedWorkflowCancelledError):
        handle.get_result()


def test_queued_workflows_endpoint(
    dbos: DBOS, skip_with_sqlite_imprecise_time: None
) -> None:
    """Test the /queues endpoint with various filters and scenarios."""

    # Set up a queue for testing
    test_queue1 = Queue("test-queue-1", concurrency=1)
    test_queue2 = Queue("test-queue-2", concurrency=1)

    @DBOS.workflow()
    def blocking_workflow(i: int) -> str:
        while True:
            time.sleep(0.1)

    # Enqueue some workflows to create queued entries
    handles: list[WorkflowHandle[str]] = []
    handles.append(test_queue1.enqueue(blocking_workflow, 1))
    handles.append(test_queue1.enqueue(blocking_workflow, 2))
    handles.append(test_queue2.enqueue(blocking_workflow, 3))

    # Test basic queued workflows endpoint
    response = requests.post("http://localhost:3001/queues", json={}, timeout=5)
    assert (
        response.status_code == 200
    ), f"Expected status 200, got {response.status_code}"

    queued_workflows = response.json()
    assert isinstance(queued_workflows, list), "Response should be a list"
    assert (
        len(queued_workflows) == 3
    ), f"Expected 3 queued workflows, got {len(queued_workflows)}"

    # Make sure it contains all the expected fields
    assert queued_workflows[0]["WorkflowName"] == blocking_workflow.__qualname__
    assert (
        queued_workflows[0]["WorkflowUUID"] == handles[0].workflow_id
    ), "Workflow ID mismatch"
    assert (
        queued_workflows[0]["Status"] == "ENQUEUED"
        or queued_workflows[0]["Status"] == "PENDING"
    )
    assert queued_workflows[0]["WorkflowClassName"] is None
    assert queued_workflows[0]["WorkflowConfigName"] is None
    assert queued_workflows[0]["AuthenticatedUser"] is None
    assert queued_workflows[0]["AssumedRole"] is None
    assert queued_workflows[0]["AuthenticatedRoles"] is None
    assert queued_workflows[0]["Input"] is None
    assert queued_workflows[0]["Output"] is None
    assert queued_workflows[0]["Error"] is None
    assert (
        queued_workflows[0]["CreatedAt"] is not None
        and len(queued_workflows[0]["CreatedAt"]) > 0
    )
    assert (
        queued_workflows[0]["UpdatedAt"] is not None
        and len(queued_workflows[0]["UpdatedAt"]) > 0
    )
    assert queued_workflows[0]["QueueName"] == test_queue1.name
    assert queued_workflows[0]["ApplicationVersion"] == DBOS.application_version
    assert queued_workflows[0]["ExecutorID"] == GlobalParams.executor_id

    # Verify sort_desc inverts the order
    filters: Dict[str, Any] = {
        "sort_desc": True,
    }
    response = requests.post("http://localhost:3001/queues", json=filters, timeout=5)
    assert response.status_code == 200
    filtered_workflows = response.json()
    assert len(filtered_workflows) == len(handles)
    assert (
        filtered_workflows[0]["WorkflowUUID"] == handles[2].workflow_id
    ), "First workflow should be the last one enqueued"

    # Only load input as requested
    filters = {
        "load_input": True,
    }
    response = requests.post("http://localhost:3001/queues", json=filters, timeout=5)
    assert response.status_code == 200
    filtered_workflows = response.json()
    assert len(filtered_workflows) == len(handles)
    assert filtered_workflows[0]["Input"] is not None
    assert "1" in filtered_workflows[0]["Input"]

    # Test all filters
    filters = {
        "workflow_name": blocking_workflow.__qualname__,
    }
    response = requests.post("http://localhost:3001/queues", json=filters, timeout=5)
    assert response.status_code == 200
    filtered_workflows = response.json()
    assert len(filtered_workflows) == len(handles)

    filters = {
        "end_time": (datetime.now(timezone.utc) - timedelta(minutes=10)).isoformat(),
    }
    response = requests.post("http://localhost:3001/queues", json=filters, timeout=5)
    assert response.status_code == 200
    filtered_workflows = response.json()
    assert len(filtered_workflows) == 0

    filters = {
        "start_time": (datetime.now(timezone.utc) + timedelta(minutes=10)).isoformat(),
    }
    response = requests.post("http://localhost:3001/queues", json=filters, timeout=5)
    assert response.status_code == 200
    filtered_workflows = response.json()
    assert len(filtered_workflows) == 0

    filters = {
        "status": ["PENDING", "ENQUEUED"],
    }
    response = requests.post("http://localhost:3001/queues", json=filters, timeout=5)
    assert response.status_code == 200
    filtered_workflows = response.json()
    assert len(filtered_workflows) == len(handles)

    filters = {
        "queue_name": test_queue1.name,
    }
    response = requests.post("http://localhost:3001/queues", json=filters, timeout=5)
    assert response.status_code == 200
    filtered_workflows = response.json()
    assert len(filtered_workflows) == 2

    filters = {"queue_name": test_queue1.name, "limit": 1, "offset": 1}
    response = requests.post("http://localhost:3001/queues", json=filters, timeout=5)
    assert response.status_code == 200
    filtered_workflows = response.json()
    assert isinstance(filtered_workflows, list), "Response should be a list"
    assert (
        len(filtered_workflows) == 1
    ), f"Expected 1 workflow, got {len(filtered_workflows)}"
    assert filtered_workflows[0]["WorkflowUUID"] == handles[1].workflow_id

    # Test with non-existent queue name
    filters = {"queue_name": "non-existent-queue"}
    response = requests.post("http://localhost:3001/queues", json=filters, timeout=5)
    assert response.status_code == 200

    empty_result = response.json()
    assert isinstance(
        empty_result, list
    ), "Response should be a list even for non-existent queue"
    assert len(empty_result) == 0, "Expected no workflows for non-existent queue"



import os
import sys

from dbos import DBOS, SetWorkflowID
from tests import client_collateral as cc
from tests.conftest import default_config

app_vers = os.environ.get("DBOS__APPVERSION")
if app_vers is None:
    DBOS.logger.error("DBOS__APPVERSION not set")
    os._exit(1)
else:
    DBOS.logger.info(f"DBOS__APPVERSION: {app_vers}")

if len(sys.argv) < 2:
    DBOS.logger.error("Usage: client_worker wfid <topic>")
    os._exit(1)

wfid = sys.argv[1]
topic = sys.argv[2] if len(sys.argv) > 2 else None

config = default_config()
DBOS(config=config)
DBOS.launch()

DBOS.logger.info(f"Starting send_test with WF ID: {wfid}")
with SetWorkflowID(wfid):
    DBOS.start_workflow(cc.send_test, topic)

os._exit(0)



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, TypedDict

import psycopg
import sqlalchemy as sa
from sqlalchemy import inspect, text
from sqlalchemy.exc import DBAPIError
from sqlalchemy.orm import Session, sessionmaker

from dbos._migration import get_sqlite_timestamp_expr

from . import _serialization
from ._error import DBOSUnexpectedStepError, DBOSWorkflowConflictIDError
from ._logger import dbos_logger
from ._schemas.application_database import ApplicationSchema
from ._sys_db import StepInfo


class TransactionResultInternal(TypedDict):
    workflow_uuid: str
    function_id: int
    output: Optional[str]  # JSON (jsonpickle)
    error: Optional[str]  # JSON (jsonpickle)
    txn_id: Optional[str]
    txn_snapshot: str
    executor_id: Optional[str]
    function_name: Optional[str]


class RecordedResult(TypedDict):
    output: Optional[str]  # JSON (jsonpickle)
    error: Optional[str]  # JSON (jsonpickle)


class ApplicationDatabase(ABC):

    def __init__(
        self,
        *,
        database_url: str,
        engine_kwargs: Dict[str, Any],
        debug_mode: bool = False,
    ):
        self.engine = self._create_engine(database_url, engine_kwargs)
        self._engine_kwargs = engine_kwargs
        self.sessionmaker = sessionmaker(bind=self.engine)
        self.debug_mode = debug_mode

    @abstractmethod
    def _create_engine(
        self, database_url: str, engine_kwargs: Dict[str, Any]
    ) -> sa.Engine:
        """Create a database engine specific to the database type."""
        pass

    @abstractmethod
    def run_migrations(self) -> None:
        """Run database migrations specific to the database type."""
        pass

    def destroy(self) -> None:
        self.engine.dispose()

    def record_transaction_output(
        self, session: Session, output: TransactionResultInternal
    ) -> None:
        try:
            session.execute(
                sa.insert(ApplicationSchema.transaction_outputs).values(
                    workflow_uuid=output["workflow_uuid"],
                    function_id=output["function_id"],
                    output=output["output"],
                    error=None,
                    txn_id="",
                    txn_snapshot=output["txn_snapshot"],
                    executor_id=(
                        output["executor_id"] if output["executor_id"] else None
                    ),
                    function_name=output["function_name"],
                )
            )
        except DBAPIError as dbapi_error:
            if self._is_unique_constraint_violation(dbapi_error):
                raise DBOSWorkflowConflictIDError(output["workflow_uuid"])
            raise

    def record_transaction_error(self, output: TransactionResultInternal) -> None:
        if self.debug_mode:
            raise Exception("called record_transaction_error in debug mode")
        try:
            with self.engine.begin() as conn:
                conn.execute(
                    sa.insert(ApplicationSchema.transaction_outputs).values(
                        workflow_uuid=output["workflow_uuid"],
                        function_id=output["function_id"],
                        output=None,
                        error=output["error"],
                        txn_id="",
                        txn_snapshot=output["txn_snapshot"],
                        executor_id=(
                            output["executor_id"] if output["executor_id"] else None
                        ),
                        function_name=output["function_name"],
                    )
                )
        except DBAPIError as dbapi_error:
            if self._is_unique_constraint_violation(dbapi_error):
                raise DBOSWorkflowConflictIDError(output["workflow_uuid"])
            raise

    @staticmethod
    def check_transaction_execution(
        session: Session, workflow_id: str, function_id: int, function_name: str
    ) -> Optional[RecordedResult]:
        rows = session.execute(
            sa.select(
                ApplicationSchema.transaction_outputs.c.output,
                ApplicationSchema.transaction_outputs.c.error,
                ApplicationSchema.transaction_outputs.c.function_name,
            ).where(
                ApplicationSchema.transaction_outputs.c.workflow_uuid == workflow_id,
                ApplicationSchema.transaction_outputs.c.function_id == function_id,
            )
        ).all()
        if len(rows) == 0:
            return None
        output, error, recorded_function_name = rows[0][0], rows[0][1], rows[0][2]
        if function_name != recorded_function_name:
            raise DBOSUnexpectedStepError(
                workflow_id=workflow_id,
                step_id=function_id,
                expected_name=function_name,
                recorded_name=recorded_function_name,
            )
        result: RecordedResult = {
            "output": output,
            "error": error,
        }
        return result

    def get_transactions(self, workflow_uuid: str) -> List[StepInfo]:
        with self.engine.begin() as conn:
            rows = conn.execute(
                sa.select(
                    ApplicationSchema.transaction_outputs.c.function_id,
                    ApplicationSchema.transaction_outputs.c.function_name,
                    ApplicationSchema.transaction_outputs.c.output,
                    ApplicationSchema.transaction_outputs.c.error,
                ).where(
                    ApplicationSchema.transaction_outputs.c.workflow_uuid
                    == workflow_uuid,
                )
            ).all()
        return [
            StepInfo(
                function_id=row[0],
                function_name=row[1],
                output=(
                    _serialization.deserialize(row[2]) if row[2] is not None else row[2]
                ),
                error=(
                    _serialization.deserialize_exception(row[3])
                    if row[3] is not None
                    else row[3]
                ),
                child_workflow_id=None,
            )
            for row in rows
        ]

    def clone_workflow_transactions(
        self, src_workflow_id: str, forked_workflow_id: str, start_step: int
    ) -> None:
        """
        Copies all steps from dbos.transctions_outputs where function_id < input function_id
        into a new workflow_uuid. Returns the new workflow_uuid.
        """

        with self.engine.begin() as conn:

            insert_stmt = sa.insert(ApplicationSchema.transaction_outputs).from_select(
                [
                    "workflow_uuid",
                    "function_id",
                    "output",
                    "error",
                    "txn_id",
                    "txn_snapshot",
                    "executor_id",
                    "function_name",
                ],
                sa.select(
                    sa.literal(forked_workflow_id).label("workflow_uuid"),
                    ApplicationSchema.transaction_outputs.c.function_id,
                    ApplicationSchema.transaction_outputs.c.output,
                    ApplicationSchema.transaction_outputs.c.error,
                    ApplicationSchema.transaction_outputs.c.txn_id,
                    ApplicationSchema.transaction_outputs.c.txn_snapshot,
                    ApplicationSchema.transaction_outputs.c.executor_id,
                    ApplicationSchema.transaction_outputs.c.function_name,
                ).where(
                    (
                        ApplicationSchema.transaction_outputs.c.workflow_uuid
                        == src_workflow_id
                    )
                    & (ApplicationSchema.transaction_outputs.c.function_id < start_step)
                ),
            )

            conn.execute(insert_stmt)

    def garbage_collect(
        self, cutoff_epoch_timestamp_ms: int, pending_workflow_ids: list[str]
    ) -> None:
        with self.engine.begin() as c:
            delete_query = sa.delete(ApplicationSchema.transaction_outputs).where(
                ApplicationSchema.transaction_outputs.c.created_at
                < cutoff_epoch_timestamp_ms
            )

            if len(pending_workflow_ids) > 0:
                delete_query = delete_query.where(
                    ~ApplicationSchema.transaction_outputs.c.workflow_uuid.in_(
                        pending_workflow_ids
                    )
                )

            c.execute(delete_query)

    @abstractmethod
    def _is_unique_constraint_violation(self, dbapi_error: DBAPIError) -> bool:
        """Check if the error is a unique constraint violation."""
        pass

    @abstractmethod
    def _is_serialization_error(self, dbapi_error: DBAPIError) -> bool:
        """Check if the error is a serialization/concurrency error."""
        pass

    @staticmethod
    def create(
        database_url: str,
        engine_kwargs: Dict[str, Any],
        schema: Optional[str],
        debug_mode: bool = False,
    ) -> "ApplicationDatabase":
        """Factory method to create the appropriate ApplicationDatabase implementation based on URL."""
        if database_url.startswith("sqlite"):
            return SQLiteApplicationDatabase(
                database_url=database_url,
                engine_kwargs=engine_kwargs,
                debug_mode=debug_mode,
            )
        else:
            # Default to PostgreSQL for postgresql://, postgres://, or other URLs
            return PostgresApplicationDatabase(
                database_url=database_url,
                engine_kwargs=engine_kwargs,
                debug_mode=debug_mode,
                schema=schema,
            )


class PostgresApplicationDatabase(ApplicationDatabase):
    """PostgreSQL-specific implementation of ApplicationDatabase."""

    def __init__(
        self,
        *,
        database_url: str,
        engine_kwargs: Dict[str, Any],
        schema: Optional[str],
        debug_mode: bool = False,
    ):
        super().__init__(
            database_url=database_url,
            engine_kwargs=engine_kwargs,
            debug_mode=debug_mode,
        )
        if schema is None:
            self.schema = "dbos"
        else:
            self.schema = schema
        ApplicationSchema.transaction_outputs.schema = schema

    def _create_engine(
        self, database_url: str, engine_kwargs: Dict[str, Any]
    ) -> sa.Engine:
        """Create a PostgreSQL engine."""
        app_db_url = sa.make_url(database_url).set(drivername="postgresql+psycopg")

        if engine_kwargs is None:
            engine_kwargs = {}

        return sa.create_engine(
            app_db_url,
            **engine_kwargs,
        )

    def run_migrations(self) -> None:
        if self.debug_mode:
            dbos_logger.warning(
                "Application database migrations are skipped in debug mode."
            )
            return
        # Check if the database exists
        app_db_url = self.engine.url
        postgres_db_engine = sa.create_engine(
            app_db_url.set(database="postgres"),
            **self._engine_kwargs,
        )
        with postgres_db_engine.connect() as conn:
            conn.execution_options(isolation_level="AUTOCOMMIT")
            if not conn.execute(
                sa.text("SELECT 1 FROM pg_database WHERE datname=:db_name"),
                parameters={"db_name": app_db_url.database},
            ).scalar():
                conn.execute(sa.text(f"CREATE DATABASE {app_db_url.database}"))
        postgres_db_engine.dispose()

        # Create the dbos schema and transaction_outputs table in the application database
        with self.engine.begin() as conn:
            # Check if schema exists first
            schema_exists = conn.execute(
                sa.text(
                    "SELECT 1 FROM information_schema.schemata WHERE schema_name = :schema_name"
                ),
                parameters={"schema_name": self.schema},
            ).scalar()

            if not schema_exists:
                schema_creation_query = sa.text(f'CREATE SCHEMA "{self.schema}"')
                conn.execute(schema_creation_query)

        inspector = inspect(self.engine)
        if not inspector.has_table("transaction_outputs", schema=self.schema):
            ApplicationSchema.metadata_obj.create_all(self.engine)
        else:
            columns = inspector.get_columns("transaction_outputs", schema=self.schema)
            column_names = [col["name"] for col in columns]

            if "function_name" not in column_names:
                # Column missing, alter table to add it
                with self.engine.connect() as conn:
                    conn.execute(
                        text(
                            f"""
                        ALTER TABLE \"{self.schema}\".transaction_outputs
                        ADD COLUMN function_name TEXT NOT NULL DEFAULT '';
                        """
                        )
                    )
                    conn.commit()

    def _is_unique_constraint_violation(self, dbapi_error: DBAPIError) -> bool:
        """Check if the error is a unique constraint violation in PostgreSQL."""
        return dbapi_error.orig.sqlstate == "23505"  # type: ignore

    def _is_serialization_error(self, dbapi_error: DBAPIError) -> bool:
        """Check if the error is a serialization/concurrency error in PostgreSQL."""
        # 40001: serialization_failure (MVCC conflict)
        # 40P01: deadlock_detected
        driver_error = dbapi_error.orig
        return (
            driver_error is not None
            and isinstance(driver_error, psycopg.OperationalError)
            and driver_error.sqlstate in ("40001", "40P01")
        )


class SQLiteApplicationDatabase(ApplicationDatabase):
    """SQLite-specific implementation of ApplicationDatabase."""

    def _create_engine(
        self, database_url: str, engine_kwargs: Dict[str, Any]
    ) -> sa.Engine:
        """Create a SQLite engine."""
        # TODO: Make the schema dynamic so this isn't needed
        ApplicationSchema.transaction_outputs.schema = None
        return sa.create_engine(database_url)

    def run_migrations(self) -> None:
        if self.debug_mode:
            dbos_logger.warning(
                "Application database migrations are skipped in debug mode."
            )
            return

        with self.engine.begin() as conn:
            # Check if table exists
            result = conn.execute(
                sa.text(
                    "SELECT name FROM sqlite_master WHERE type='table' AND name='transaction_outputs'"
                )
            ).fetchone()

            if result is None:
                conn.execute(
                    sa.text(
                        f"""
                        CREATE TABLE transaction_outputs (
                            workflow_uuid TEXT NOT NULL,
                            function_id INTEGER NOT NULL,
                            output TEXT,
                            error TEXT,
                            txn_id TEXT,
                            txn_snapshot TEXT NOT NULL,
                            executor_id TEXT,
                            function_name TEXT NOT NULL DEFAULT '',
                            created_at BIGINT NOT NULL DEFAULT {get_sqlite_timestamp_expr()},
                            PRIMARY KEY (workflow_uuid, function_id)
                        )
                        """
                    )
                )
                conn.execute(
                    sa.text(
                        "CREATE INDEX transaction_outputs_created_at_index ON transaction_outputs (created_at)"
                    )
                )

    def _is_unique_constraint_violation(self, dbapi_error: DBAPIError) -> bool:
        """Check if the error is a unique constraint violation in SQLite."""
        return "UNIQUE constraint failed" in str(dbapi_error.orig)

    def _is_serialization_error(self, dbapi_error: DBAPIError) -> bool:
        """Check if the error is a serialization/concurrency error in SQLite."""
        # SQLite database is locked or busy errors
        error_msg = str(dbapi_error.orig).lower()
        return (
            "database is locked" in error_msg or "database table is locked" in error_msg
        )



from dbos import DBOS, DBOSConfiguredInstance


@DBOS.dbos_class()
class DBOSTestRegDup(DBOSConfiguredInstance):
    """DBOSTestRegDup duplicates the name of a class defined dupname_classdefs1.py"""

    def __init__(self, instance_name: str) -> None:
        super().__init__(instance_name)
